#ifndef _FL_DFB_WindowManager_H__
#define _FL_DFB_WindowManager_H__

#include <directfb.h>
#include "FL_DFB_Window.h"

#include <vector>
using std::vector;

#ifdef VA_PLATFORM_EMMA3PF
    #define ENABLE_IPC_INPUT 1 //enable ipc input on emma3pf
#else
    #define ENABLE_IPC_INPUT 0
#endif

class FL_DFB_WindowManager {
public:    
    static const unsigned int INVALID_INDEX = 0xFFFFFFFF;
    static const unsigned int MAX_XID = 0xFFFF;
    
public:
    FL_DFB_WindowManager();
    virtual ~FL_DFB_WindowManager();

    static FL_DFB_WindowManager* getInstance();

    bool getLayer(IDirectFBDisplayLayer** layer);
    bool createWindowSurface(int w, int h, IDirectFBSurface** surface);
    bool destroyWindowSurface(IDirectFBSurface* surface);

    bool addWindow(FL_DFB_Window* window);
    bool removeWindow(FL_DFB_Window* window);

    bool raiseToTop(FL_DFB_Window* window);
    bool lowerToBottom(FL_DFB_Window* window);

    bool setWindowOpacity(FL_DFB_Window* window, unsigned char old);
    bool setWindowPosition(FL_DFB_Window* window, int dx, int dy);
    bool setWindowSize(FL_DFB_Window* window, int ow, int oh);

    FL_DFB_Window* windowAt(int x, int y);

    bool requestWindowFocus(FL_DFB_Window* window);
    bool grabPointer(FL_DFB_Window*);
    bool ungrabPointer(FL_DFB_Window*);
    bool grabKeyboard(FL_DFB_Window*);
    bool ungrabKeyboard(FL_DFB_Window*);

    unsigned int generateXID();
    FL_DFB_Window* findWindow(unsigned int xid);
    bool updateWindow(FL_DFB_Window* window, DFBRegion* region = NULL, bool refresh = true);
        
    //event    
    void do_queued_events();
#if ENABLE_IPC_INPUT    
    void do_ipc_queued_events();
#endif    

    void postEventToWM(DFBEvent* event);

    FL_DFB_Window* getKeyboardWindow();
    FL_DFB_Window* getPointerWindow(int x, int y);

private:
    IDirectFB* mDFB;
    IDirectFBDisplayLayer* mLayer;
    IDirectFBSurface* mLayerSurface;
    DFBSurfacePixelFormat mLayerPixelFormat;
    int mLayerW;
    int mLayerH;
    unsigned int mGID;
    vector<FL_DFB_Window*> mWindows;

    unsigned int windowIndex(FL_DFB_Window* window);
    void updateFocus();
    void switchFocus(FL_DFB_Window* to);
    void processMotion(int dx, int dy);
    void processInput(DFBInputEvent& event);
    void sendWindowEvent(FL_DFB_Window* w, DFBWindowEvent& event);
    void sendWindowEventToFLTK(DFBWindowEvent& event);
    void withdrawWindow(FL_DFB_Window* window);    

    void drawWindow(int winNum, DFBRegion* updateRegion);    
    void updateRegions(int winNum,int x1,int y1,int x2,int y2); 

    //event
    IDirectFBEventBuffer* mEventBuffer;
    int mEventFD;
#if ENABLE_IPC_INPUT    
    pthread_t mIPCThreadID;
    bool mIPCExit;
#endif  

    int mButtons;
    int mModifiers;
    int mLocks;
    int mCursorX;
    int mCursorY;
    bool mSoftwareCursor;

    FL_DFB_Window* mFocusedWindow;
    FL_DFB_Window* mKeyboardWindow; //for grab
    FL_DFB_Window* mPointerWindow;  // for grab
    FL_DFB_Window* mEnteredWindow;
    bool mScaledTo1920x1080; //this mode is for Sharp's DFB
};

#endif//_FL_DFB_WindowManager_H__

