#ifndef __FLTK_DFB_EX_H__
#define __FLTK_DFB_EX_H__ 

/************************************************************************
 * External FLTK`s functions which available in DirectFB`s version of the library only.
*/

/* create offscreen compatible with primary surface */
extern void* fl_create_offscreen(int w,int h);

/* create ARGB offscreen*/
extern void* fl_create_offscreen_rgb(int w,int h);

/* delete offscreen*/
extern void fl_delete_offscreen(void* surface);

/* set offscreen as current drawing surface*/
extern void fl_begin_offscreen(void* surface );

/* restore old drawing surface*/
extern void fl_end_offscreen();

/* copy from offscreen to current surface*/
extern void fl_copy_offscreen(int _x,int _y,int _w,int _h,void* offscreen,int srcx,int srcy);

/* bkcolor color is excluded when copying */
extern void fl_copy_offscreen_bk(int _x,int _y,int _w,int _h,void* offscreen,int srcx,int srcy,ulong bkcolor);

/* 
 *  filter and smooth functions. 
 *  They draw good-looking image especially if your display has low resolution ( e.c.640x480 ).
 */
 
/* available masks for filtering  
 * 
 * 1 - 2 X 2		 1 1
 * 					 1 1  
 * 
 * 2 - 3 X 3        1 1 1 
 *                  1 1 1 
 *                  1 1 1
 *
 * 3 - 3 X 3        1 1 1
 *                  1 2 1
 *                  1 1 1 
 * 
 * 4 - 3 X 3        1 2 1
 *                  2 4 2
 *                  1 2 1 
 * 
 * 5 - 3 X 3        0 2 0
 *                  2 4 2
 *                  0 2 0
 * 
 * 6 - 4 X 4       1 1 1 1 
 *                 1 1 1 1
 *                 1 1 1 1 
 *                 1 1 1 1
 * 
 * 7 - 7 X 7   1 2  3  4  3 2 1 
 *             2 4  6  8  6 4 2 
 *             3 6  9 12  9 6 5
 *             4 8 12 16 12 8 4 
 *             3 6  9 12  9 6 5 
 *             2 4  6  8  6 4 2 
 *             1 2  3  4  3 2 1
 */
/* filter offscreen rectangle using mask */
extern void fl_filter_offscreen(void* offscreen,int x,int y,int w,int h,int mask);

/* filter widget's rectangle*/
extern void fl_filter_widget(Fl_Widget *wdg, int mask ); 

/* draw a smooth line */
extern void fl_line(double x1, double y1, double x2, double y2);

/* draw a smooth edge. 
 * The egde is a line having only one smooth side ( left hand side ). */
extern void fl_edge(double x1, double y1, double x2, double y2);

/* start drawing smooth lines */
extern void fl_begin_smooth_line();

/* start drawing a closed sequence of smooth lines*/
extern void fl_begin_smooth_loop();

/*
 *  start drawing of polygon with smooth edges
 *  list the points in clockwise direction
 1         2
  *--->-----* 
  |######### \
  |########## \
  ^########### V
  |############ \
  |############# \
  *-------<-------* 
 0                 3
*/
extern void fl_begin_smooth_polygon();

/* set and get a mask which be used after fl_draw() function*/
extern void set_text_smooth(int);
extern int get_text_smooth();

/* hide/show the cursor*/
extern void fl_show_cursor(int);


/*******************************************************************
 *   it's an example of using of smooth drawing 
 *   Fl_Smooth_Clock class declaration
 *   
 */

#include <FL/Fl_Clock.H>

class Fl_Smooth_Clock: public Fl_Clock{
public: 
    Fl_Smooth_Clock(int x,int y,int w,int h, const char *l = 0):Fl_Clock(x,y,w,h,l){};    
protected:    
    void drawsmhands(Fl_Color fill, Fl_Color line);
    virtual void draw();
    void draw(int X, int Y, int W, int H); 
};


/****************************************************************************/
#endif
