
#include "FL/FL_DFB_Window.h"
#include "FL/FL_DFB_WindowManager.h"

FL_DFB_Window::FL_DFB_Window(int x, int y, int w, int h)
    : mWM(NULL), mX(x), mY(y), mW(w), mH(h),
    mOpacity(0), mSurface(NULL)
{
    mWM = FL_DFB_WindowManager::getInstance();
    mXID = mWM->generateXID();
    mWM->createWindowSurface(w, h, &mSurface);
    mWM->addWindow(this);
}

FL_DFB_Window::~FL_DFB_Window()
{
    mWM->removeWindow(this);
    mWM->destroyWindowSurface(mSurface);
    mSurface = NULL;
}

bool FL_DFB_Window::getSurface(IDirectFBSurface** ret_surface)
{
    mSurface->AddRef(mSurface);
    *ret_surface = mSurface;
    return true;
}

bool FL_DFB_Window::setOpacity(unsigned char opacity)
{
    int o = mOpacity;
    if (mOpacity == opacity)
        return false;
    mOpacity = opacity;
    return mWM->setWindowOpacity(this, o);
}

bool FL_DFB_Window::moveTo(int x, int y)
{
    if (x == mX && y == mY)
        return false;
    int dx = x - mX;
    int dy = y - mY;
    mX = x;
    mY = y;
    return mWM->setWindowPosition(this, dx, dy);
}

bool FL_DFB_Window::resize(int w, int h)
{
    if (w == mW && h == mH)
        return false;

    mWM->destroyWindowSurface(mSurface);
    mSurface = NULL;
    mWM->createWindowSurface(w, h, &mSurface);

    int ow = mW;
    int oh = mH;
    mW = w;
    mH = h;

    return mWM->setWindowSize(this, ow, oh);
}

bool FL_DFB_Window::setBounds(int x, int y, int w, int h)
{
    //TODO
    moveTo(x, y);
    resize(w, h);
    return true;
}

bool FL_DFB_Window::raiseToTop()
{
    return mWM->raiseToTop(this);
}

bool FL_DFB_Window::lowerToBottom()
{
    return mWM->lowerToBottom(this);
}

bool FL_DFB_Window::requestFocus()
{
    return mWM->requestWindowFocus(this);
}

bool FL_DFB_Window::grabPointer()
{
    return mWM->grabPointer(this);    
}

bool FL_DFB_Window::ungrabPointer()
{
    return mWM->ungrabPointer(this);
}

bool FL_DFB_Window::grabKeyboard()
{
    return mWM->grabKeyboard(this);
}

bool FL_DFB_Window::ungrabKeyboard()
{
    return mWM->ungrabKeyboard(this);
}

