#include "FL/x.H"
#include "FL/Fl.H"
#include "FL/FL_DFB_WindowManager.h"
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/time.h>
#include <poll.h>
#include <directfb_keyboard.h>
#include <directfb_keynames.h>

#include <assert.h>

//#define D_TRACE(...) do {printf("%s %d: ", __FUNCTION__, __LINE__);printf(__VA_ARGS__);} while(0)
#define D_TRACE(...) do { } while(0)

#if ENABLE_IPC_INPUT
#include <sys/msg.h>
#include <pthread.h>

//just copy of mile's ipc static library to get ride of the dependency of it.
extern "C" {
    enum CMsgTypes
    {
        MSG_TYP_Unknow,
        MSG_TYP_INPUTKEY_CMD = 1,
        MSG_TYP_USB_CMD,
        MSG_TYP_MAINUI_CMD,
        MSG_TYP_BROWSER_CMD,
        MSG_TYP_MEDIA_CMD,

    };
    typedef struct
    {
    	long	cmd_id;
    	short	cmd_param1_num;
    	short	cmd_param2_num;
    	long	cmd_param1[32];
    	char	cmd_param2[8][256];
    }SIpcCmdInf;
    
    typedef struct
    {
    	long					m_type;
    	
    	union
    	{
    		SIpcCmdInf		m_cmd_inf;
    		DFBInputEvent m_key_cod;
    	}__attribute__((__packed__));
    	
    }SIpcMsgInf;
    static const key_t MSG_QUEUE_KEY = 0x20000001;
}
////////////////////
#endif

static FL_DFB_WindowManager* gWM = NULL;

static void fd_callback(int fd, void *)
{
    if (gWM)
        gWM->do_queued_events();
}

#if ENABLE_IPC_INPUT
static void* ipcThread(void*)
{
    if (gWM)
        gWM->do_ipc_queued_events();
    return 0;
}
#endif

#if 0
static const DirectFBKeySymbolNames(keynames);
static const DirectFBKeyIdentifierNames(idnames);

static int compare_symbol(const void *a, const void *b)
{
    DFBInputDeviceKeySymbol *symbol = (DFBInputDeviceKeySymbol*) a;
    struct DFBKeySymbolName *symname = (struct DFBKeySymbolName*) b;

    return *symbol - symname->symbol;
}

static int
compare_id( const void *a, const void *b ) {
     DFBInputDeviceKeyIdentifier      *id     = (DFBInputDeviceKeyIdentifier *) a;
     struct DFBKeyIdentifierName *idname = (struct DFBKeyIdentifierName *) b;

     return *id - idname->identifier;
}

static void show_key_event(DFBInputEvent& evt)
{
    struct DFBKeySymbolName *symbol_name;
    struct DFBKeyIdentifierName *id_name;

    symbol_name = (struct DFBKeySymbolName*)bsearch( &evt.key_symbol, keynames,
            sizeof(keynames) / sizeof(keynames[0]) - 1,
            sizeof(keynames[0]), compare_symbol);

    id_name = (struct DFBKeyIdentifierName*)bsearch( &evt.key_id, idnames,
            sizeof(idnames) / sizeof(idnames[0]) - 1,
            sizeof(idnames[0]), compare_id );

    printf("key_symbol: 0x%X ", evt.key_symbol);
    if (symbol_name) {
        printf(":%s ", symbol_name->name);
    }
    printf("key_code: %d ", evt.key_code);
    if (id_name) {
        printf("%s ", id_name->name);
    }
    printf(" key_id: %d", evt.key_id);
    printf("\n");
}
#endif

FL_DFB_WindowManager* FL_DFB_WindowManager::getInstance()
{
    return gWM;
}

FL_DFB_WindowManager::FL_DFB_WindowManager()
{
    gWM = this;

    mDFB = fl_dfb;
    mLayer = fl_display_layer;
    mLayerSurface = fl_screen->surface;
    DFBDisplayLayerConfig layer_config;
    DFBCHECK(mLayer->GetConfiguration(mLayer, &layer_config));
    mLayerPixelFormat = layer_config.pixelformat;
    mLayerW = layer_config.width;
    mLayerH = layer_config.height;

    mGID = 0;
    DFBCHECK(mDFB->CreateInputEventBuffer(mDFB, DICAPS_ALL, DFB_TRUE, &mEventBuffer));
    DFBCHECK(mEventBuffer->CreateFileDescriptor(mEventBuffer, &mEventFD));
    mButtons = 0;
    mModifiers = 0;
    mLocks = 0;    
    mCursorX = mLayerW / 2;
    mCursorY = mLayerH / 2;
    mSoftwareCursor = false;

    mFocusedWindow = NULL;
    mKeyboardWindow = NULL;
    mPointerWindow = NULL;
    mEnteredWindow = NULL;

    mScaledTo1920x1080 = !!getenv("VA_WEBKIT_SCALED_TO_1920x1080"); //this mode is for Sharp's DirectFB
    if (mLayerW != 1920 || mLayerH != 1080)
        mScaledTo1920x1080 = false;

    Fl::add_fd(mEventFD, POLLIN, fd_callback);

#if ENABLE_IPC_INPUT
    mIPCThreadID = 0;
    mIPCExit = false;
    pthread_create(&mIPCThreadID, NULL, ipcThread, NULL);
#endif
}

FL_DFB_WindowManager::~FL_DFB_WindowManager()
{
#if ENABLE_IPC_INPUT
    if (mIPCThreadID) {
        mIPCExit = true;
        //pthread_cancel(mIPCThreadID);
        pthread_join(mIPCThreadID, NULL);
        mIPCThreadID = 0;
    }
#endif
    
    for(unsigned int i = 0; i < mWindows.size(); i++) {
        D_TRACE("lived windows @%p\n", mWindows.at(i));
    }

    if (mEventBuffer) {
        mEventBuffer->Release(mEventBuffer);
        mEventBuffer = NULL;
        mEventFD = -1;
    }

    gWM = NULL;    
}

unsigned int FL_DFB_WindowManager::generateXID()
{
    do {
        bool conflict = false;
        mGID = (mGID + 1) % MAX_XID;
        for(unsigned int i = 0; i < mWindows.size(); i++)
            if (mWindows.at(i)->getXID() == mGID) {
                conflict = true;
                break;
            }
        if (!conflict)
            return mGID;
    } while(1);
}

FL_DFB_Window* FL_DFB_WindowManager::findWindow(unsigned int xid)
{
    for(unsigned int i = 0; i < mWindows.size(); i++)
        if (mWindows.at(i)->getXID() == xid)
            return mWindows.at(i);   
    return NULL;
}

bool FL_DFB_WindowManager::getLayer(IDirectFBDisplayLayer** layer)
{
    *layer = mLayer;
    return true;
}

bool FL_DFB_WindowManager::createWindowSurface(int w, int h, IDirectFBSurface** surface)
{
    if (w <= 0 || h <= 0 || surface == NULL) {
        D_TRACE("invalid param w = %d, h = %d, surface = %p\n", w, h, surface);
        return false;
    }
    DFBSurfaceDescription desc;
    DFBResult ret;
    desc.flags = (DFBSurfaceDescriptionFlags)(DSDESC_WIDTH | DSDESC_HEIGHT | DSDESC_PIXELFORMAT /*| DSDESC_CAPS*/);
    desc.width = w;
    desc.height = h;
    desc.pixelformat = mLayerPixelFormat;
#ifndef VA_PLATFORM_X86
    desc.flags = (DFBSurfaceDescriptionFlags)(desc.flags | DSDESC_CAPS);
    desc.caps = (DFBSurfaceCapabilities)(DSCAPS_VIDEOONLY);
#endif
    DFBCHECK((ret = mDFB->CreateSurface(mDFB, &desc, surface)));
    return (ret == DFB_OK);
}

bool FL_DFB_WindowManager::destroyWindowSurface(IDirectFBSurface* surface)
{
    if (surface)
        surface->Release(surface);
    return true;
}

unsigned int FL_DFB_WindowManager::windowIndex(FL_DFB_Window* window)
{
    for(unsigned int i = 0; i < mWindows.size(); i++) {
        if (mWindows.at(i) == window)
            return i;
    }
    return INVALID_INDEX;
}

bool FL_DFB_WindowManager::addWindow(FL_DFB_Window* window)
{
    if (!window || windowIndex(window) != INVALID_INDEX) {
        D_TRACE("window (%p) has been in window list\n", window);
        return false;
    }
    D_TRACE("add window pointer %p\n", window);
    mWindows.push_back(window);
    if (window->isShown())
        updateWindow(window, NULL, true);
    return true;
}

bool FL_DFB_WindowManager::removeWindow(FL_DFB_Window* window)
{
    unsigned int i = windowIndex(window);
    if (i == INVALID_INDEX) {
        D_TRACE("invalid window pointer %p\n", window);
        return false;
    }
    D_TRACE("remove window %p at %d\n", window, i);
    if (window->isShown())
        window->setOpacity(0);
    mWindows.erase(mWindows.begin() + i);
    return true;
}

bool FL_DFB_WindowManager::raiseToTop(FL_DFB_Window* window)
{
    unsigned int i = windowIndex(window);
    if (i == INVALID_INDEX) {
        D_TRACE("invalid window pointer %p\n", window);
        return false;
    }
    if (i < mWindows.size() - 1) {
        mWindows.erase(mWindows.begin() + i);
        mWindows.push_back(window);
        updateWindow(window, NULL, true);
    }
    return true;
}

bool FL_DFB_WindowManager::lowerToBottom(FL_DFB_Window* window)
{ 
    unsigned int i = windowIndex(window);
    if (i == INVALID_INDEX) {
        D_TRACE("invalid window pointer %p\n", window);
        return false;
    }
    if (i > 0) {
        mWindows.erase(mWindows.begin() + i);
        mWindows.insert(mWindows.begin(), window);
        updateWindow(window, NULL, true);
    }
    return true;
}

bool FL_DFB_WindowManager::setWindowOpacity(FL_DFB_Window* window, unsigned char old)
{
    bool show = !old && window->mOpacity;
    bool hide = old && !window->mOpacity;

    updateWindow(window, NULL, true);

    if (show || hide)
        updateFocus();
    if (hide)
        withdrawWindow(window);    

    return true;
}

bool FL_DFB_WindowManager::setWindowPosition(FL_DFB_Window* window, int dx, int dy)
{
    unsigned int i = windowIndex(window);
    if (i == INVALID_INDEX) {
        D_TRACE("invalid window pointer %p\n", window);
        return false;
    }

    if (window->isShown()) {
        DFBRegion region = {0, 0, window->mW, window->mH};
        if (dx > 0)
            region.x1 -= dx;
        else
            region.x2 -= dx;
        if (dy > 0)
            region.y1 -= dy;
        else
            region.y2 -= dy;
        updateWindow(window, &region, true);
    }

    DFBWindowEvent we;
    we.type = DWET_POSITION;
    we.x = window->x();
    we.y = window->y();
    sendWindowEvent(window, we);

    return true;
}

bool FL_DFB_WindowManager::setWindowSize(FL_DFB_Window* window, int ow, int oh)
{
    unsigned int i = windowIndex(window);
    if (i == INVALID_INDEX) {
        D_TRACE("invalid window pointer %p\n", window);
        return false;
    }
    if (window->isShown()) {
        if (ow > window->mW) {
            DFBRegion region = {window->mW, 0, ow - 1, MIN(window->mH, oh) - 1};
            updateWindow(window, &region, (oh <= window->mH) ? true : false);
        }
        if (oh > window->mH) {
            DFBRegion region = {0, window->mH, MIN(window->mW, ow) - 1, oh - 1};
            updateWindow(window, &region, true);
        }
    }

    DFBWindowEvent we;
    we.type = DWET_SIZE;
    we.w = window->w();
    we.h = window->h();
    sendWindowEvent(window, we);

    updateFocus();

    return true;
}

void FL_DFB_WindowManager::drawWindow(int winNum, DFBRegion* updateRegion)
{
	FL_DFB_Window* currentWindow = mWindows.at(winNum);
    int x = currentWindow->mX;
    int y = currentWindow->mY;
    int w = currentWindow->mW;
    int h = currentWindow->mH;

    if (!currentWindow->isShown() || !currentWindow->mSurface)
    	return;
    
    DFBRegion currentRegion = DFB_REGION_INIT_INTERSECTED(updateRegion, x, y, x + w - 1, y + h - 1);
    if (currentRegion.x1 >= currentRegion.x2 || currentRegion.y1 >= currentRegion.y2)
    	return;
    
    DFBRectangle src = DFB_RECTANGLE_INIT_FROM_REGION(&currentRegion);
    dfb_rectangle_translate(&src, -x, -y);
    
    DFBCHECK(mLayerSurface->SetPorterDuff(mLayerSurface, DSPD_SRC_OVER));
    
    if (currentWindow->mOpacity != 0xFF) {
    	DFBCHECK(mLayerSurface->SetBlittingFlags(mLayerSurface, (DFBSurfaceBlittingFlags)(DSBLIT_BLEND_ALPHACHANNEL | DSBLIT_BLEND_COLORALPHA)));
        DFBCHECK(mLayerSurface->SetColor(mLayerSurface, 0, 0, 0, currentWindow->mOpacity));
    } else {
        DFBCHECK(mLayerSurface->SetBlittingFlags(mLayerSurface, DSBLIT_BLEND_ALPHACHANNEL));
    }

    DFBCHECK(mLayerSurface->SetPorterDuff(mLayerSurface, DSPD_NONE));
    DFBCHECK(mLayerSurface->SetBlittingFlags(mLayerSurface, DSBLIT_NOFX));
    
    D_TRACE("blit {%d, %d, %d, %d} --> {%d, %d}\n", src.x, src.y, src.w, src.h, currentRegion.x1, currentRegion.y1);
    
    if (mScaledTo1920x1080) {
    	DFBRectangle dst;
    	dst.x = currentRegion.x1 * 2;
    	dst.y = currentRegion.y1 * 2;
    	dst.w = src.w * 2;
    	dst.h = src.h * 2;
    	DFBCHECK(mLayerSurface->StretchBlit(mLayerSurface, currentWindow->mSurface, &src, &dst));
    } else {
    	DFBCHECK(mLayerSurface->Blit(mLayerSurface, currentWindow->mSurface, &src, currentRegion.x1, currentRegion.y1));
    }
}

void FL_DFB_WindowManager::updateRegions(int winNum,int x1,int y1,int x2,int y2)
{
	int       i      = winNum; //start
	DFBRegion region = { x1, y1, x2, y2 };
	   
	assert(i >= 0);
	assert(x1 <= x2);
	assert(y1 <= y2);
	
	/* Find next intersecting window. */
	while (i >= 0)
	{
		FL_DFB_Window* currentWindow = mWindows.at(i);
		
		if(currentWindow->isShown())
		{
			int x = currentWindow->mX;
			int y = currentWindow->mY;
			int w = currentWindow->mW;
			int h = currentWindow->mH;
			//DFBRegion curRegion = {x, y, x + w, y + h};
			
			if(dfb_region_intersect(&region, x, y, x + w - 1, y + h - 1))
			{
				D_TRACE("[FL_DFB_WindowManager] Find instersect window[%d]\n",i);
				break;
			}
		}
		
        i--;
    }
    
    if(i >= 0)
    {
    	FL_DFB_Window* window = mWindows.at(i);
    	
    	if(window->mOpacity < 0xFF) 
    	{
    		/* draw everything below */
    		updateRegions(i-1, x1, y1, x2, y2 );
    	}else {
    		/* left */
    		if (region.x1 != x1)
    			updateRegions(i - 1, x1, region.y1, region.x1 - 1, region.y2 );
    		
    		/* upper */
    		if (region.y1 != y1)
    			updateRegions(i - 1, x1, y1, x2, region.y1 - 1 );
    		
    		/* right */
    		if (region.x2 != x2)
    			updateRegions(i - 1, region.x2 + 1, region.y1, x2, region.y2 );
    		
    		/* lower */
    		if (region.y2 != y2)
    			updateRegions(i - 1, x1, region.y2 + 1, x2, y2 );
    	}
    	
    	drawWindow( i, &region);   	
    } else {
    	//draw black background
    	DFBCHECK(mLayerSurface->SetColor(mLayerSurface, 0,0,0,0)); 
    	DFBCHECK(mLayerSurface->SetPorterDuff(mLayerSurface, DSPD_NONE));
    	DFBCHECK(mLayerSurface->SetDrawingFlags(mLayerSurface, DSDRAW_NOFX));

    	if(mScaledTo1920x1080){
    	    DFBCHECK(mLayerSurface->FillRectangle(mLayerSurface, x1 * 2, y1 * 2 , (x2 - x1 + 1) * 2, (y2 - y1 + 1) * 2));
    	}else{
    	    DFBCHECK(mLayerSurface->FillRectangle(mLayerSurface, x1, y1, x2 - x1 + 1, y2 - y1 + 1));
    	}
    }
}

bool FL_DFB_WindowManager::updateWindow(FL_DFB_Window* window, DFBRegion* region, bool refresh)
{
    //TODO 
    // 1, use batch blit on emma3pf
    // 2, implement software mouse
    // 3, optimize according window setting.
    //unsigned int i = windowIndex(window);
    //if (i == INVALID_INDEX) {
    //    D_TRACE("invalid window pointer %p\n", window);
    //    return false;
    //}
    
    DFBRegion updateRegion;    
    if (region)
        updateRegion = DFB_REGION_INIT_TRANSLATED(region, (window ? window->mX : 0), (window ? window->mY : 0));    	
    else
        updateRegion = DFB_REGION_INIT_FROM_RECTANGLE_VALS(
        	(window ? window->mX : 0), (window ? window->mY : 0), (window ? window->mW : mLayerW), (window ? window->mH : mLayerH));
	
    dfb_region_clip(&updateRegion, 0, 0, mLayerW - 1, mLayerH -1);
    if (updateRegion.x1 >= updateRegion.x2 || updateRegion.y1 >= updateRegion.y2)
        return true;

    D_TRACE("%p, {%d, %d, %d, %d}, %s\n", window, updateRegion.x1, updateRegion.y1, updateRegion.x2, updateRegion.y2, refresh ? "refresh" : "no refresh");

    if (mScaledTo1920x1080) {
        DFBRegion region;
        region.x1 = updateRegion.x1 * 2;
        region.y1 = updateRegion.y1 * 2;
        region.x2 = updateRegion.x2 * 2;
        region.y2 = updateRegion.y2 * 2;
        DFBCHECK(mLayerSurface->SetClip(mLayerSurface, &region));
    } else {
        DFBCHECK(mLayerSurface->SetClip(mLayerSurface, &updateRegion));
    }

    //FIXME current webit window is fullscreen and opacity
    //if(mWindows.size() == 0 || !mWindows.at(0)->isShown())
    //    DFBCHECK(mLayerSurface->Clear(mLayerSurface, 0, 0, 0, 0));
        
    updateRegions(mWindows.size() - 1, updateRegion.x1, updateRegion.y1, updateRegion.x2, updateRegion.y2);

    if (refresh) {
        if (mScaledTo1920x1080) {
            DFBRegion region;
            region.x1 = updateRegion.x1 * 2;
            region.y1 = updateRegion.y1 * 2;
            region.x2 = updateRegion.x2 * 2;
            region.y2 = updateRegion.y2 * 2;
            DFBCHECK(mLayerSurface->Flip(mLayerSurface, &region, DSFLIP_BLIT));
        } else        
            DFBCHECK(mLayerSurface->Flip(mLayerSurface, &updateRegion, DSFLIP_BLIT));    
    }

    DFBCHECK(mLayerSurface->SetClip(mLayerSurface, NULL));

    return true;
}

FL_DFB_Window* FL_DFB_WindowManager::windowAt(int x, int y)
{
    for(int i = mWindows.size() - 1; i >= 0; i--) {
        FL_DFB_Window* w = mWindows.at(i);
        if (!w->isShown())
            continue;
        int x1 = w->x();
        int y1 = w->y();
        int x2 = w->x() + w->w() - 1;
        int y2 = w->y() + w->h() - 1;
        if (x1 <= x && x2 >= x && y1 <= y && y2 >= y)
            return w;
    }
    return NULL;
}

bool FL_DFB_WindowManager::requestWindowFocus(FL_DFB_Window* window)
{
    switchFocus(window);

    FL_DFB_Window* entered = mEnteredWindow;
    if (entered && entered != window) {
        DFBWindowEvent we;

        mEnteredWindow = NULL;
        we.type = DWET_LEAVE;
        we.x = mCursorX - entered->x();
        we.y = mCursorY - entered->y();
        sendWindowEvent(entered, we);
    }
    return (mFocusedWindow == window);
}

bool FL_DFB_WindowManager::grabPointer(FL_DFB_Window* window)
{
    if (mPointerWindow) {
        D_TRACE("pointer has been grabbed by window %p\n", window);
        return false;
    }
    mPointerWindow = window;
    return true;
}

bool FL_DFB_WindowManager::ungrabPointer(FL_DFB_Window* window)
{
    if (mPointerWindow != window) {
        D_TRACE("pointer is not grabbed by window %p\n", window);
        return false;
    }
    mPointerWindow = NULL;
    return true;
}

bool FL_DFB_WindowManager::grabKeyboard(FL_DFB_Window* window)
{
    if (mKeyboardWindow) {
        D_TRACE("keyboard has been grabbed by window %p\n", window);
        return false;
    }
    mKeyboardWindow = window;
    return true;
}

bool FL_DFB_WindowManager::ungrabKeyboard(FL_DFB_Window* window)
{
    if (mKeyboardWindow != window) {
        D_TRACE("keyboard is not grabbed by window %p\n", window);
        return false;
    }
    mKeyboardWindow = NULL;
    return true;
}

FL_DFB_Window* FL_DFB_WindowManager::getKeyboardWindow()
{
    if (mKeyboardWindow)
        return mKeyboardWindow;
    if (mFocusedWindow)
        return mFocusedWindow;
    if (mWindows.size() > 0)
        return mWindows.at(mWindows.size() - 1);
    return NULL;
}

FL_DFB_Window* FL_DFB_WindowManager::getPointerWindow(int x, int y)
{
    if (mPointerWindow)
        return mPointerWindow;    
    if (x < 0)
        x = mCursorX;
    if (y < 0)
        y = mCursorY;
    return windowAt(x, y);
}

void FL_DFB_WindowManager::withdrawWindow(FL_DFB_Window* window)
{
    if (mFocusedWindow == window)
        mFocusedWindow = NULL;
    if (mKeyboardWindow == window)
        mKeyboardWindow = NULL;
    if (mPointerWindow == window)
        mPointerWindow = NULL;
    if (mEnteredWindow == window)
        mEnteredWindow = NULL;
}

void FL_DFB_WindowManager::sendWindowEvent(FL_DFB_Window* w, DFBWindowEvent& event)
{
    event.clazz = DFEC_WINDOW;
    event.window_id = w->getXID();
    event.cx = mCursorX;
    event.cy = mCursorY;
    event.buttons = (DFBInputDeviceButtonMask)mButtons;
    event.modifiers = (DFBInputDeviceModifierMask)mModifiers;
    event.locks = (DFBInputDeviceLockState)mLocks;

    gettimeofday(&event.timestamp, NULL);
    D_TRACE("%s: w = %p\n", __FUNCTION__, w);
    sendWindowEventToFLTK(event);
}

void FL_DFB_WindowManager::postEventToWM(DFBEvent* event)
{
    if (event && mEventBuffer) {
        D_TRACE("got input event from ipc\n");
        mEventBuffer->PostEvent(mEventBuffer, event);
    }
}

extern IDirectFBEventBuffer* fl_dfb_event_buffer;
void FL_DFB_WindowManager::sendWindowEventToFLTK(DFBWindowEvent& event)
{
    //TODO 
    // just to call FLTK fl::handle_event ?????
    if (fl_dfb_event_buffer) {
        DFBEvent e;
        memcpy(&e, &event, sizeof(DFBWindowEvent));
        fl_dfb_event_buffer->PostEvent(fl_dfb_event_buffer, &e);
    }
}

void FL_DFB_WindowManager::switchFocus(FL_DFB_Window* to)
{
    DFBWindowEvent we;
    FL_DFB_Window* from = mFocusedWindow;

    if (from == to)
        return;

    mFocusedWindow = to;

    if (from) {
        we.type = DWET_LOSTFOCUS;
        sendWindowEvent(from, we);
    }
    if (to) {
        we.type = DWET_GOTFOCUS;
        sendWindowEvent(to, we);
    }
}

void FL_DFB_WindowManager::updateFocus()
{
    if (mPointerWindow)
        return;
    FL_DFB_Window* before = mEnteredWindow;
    FL_DFB_Window* after = windowAt(mCursorX, mCursorY);

    mEnteredWindow = after;
    if (after != before) {
        DFBWindowEvent we;

        if (before) {
            we.type = DWET_LEAVE;
            we.x = mCursorX - before->x();
            we.y = mCursorY - before->y();

            sendWindowEvent(before, we);
        }

        switchFocus(after);

        if (after) {
            we.type = DWET_ENTER;
            we.x = mCursorX - after->x();
            we.y = mCursorY - after->y();

            sendWindowEvent(after, we);
        }
    }
}

void FL_DFB_WindowManager::processMotion(int dx, int dy)
{
    int new_cx, new_cy;
    DFBWindowEvent we;
   
    if (mSoftwareCursor) {
        new_cx = MIN(mLayerW, (mCursorX + dx));
        new_cy = MIN(mLayerH, (mCursorY + dy));
        new_cx = MAX(0, new_cx);
        new_cy = MAX(0, new_cy);
    } else {
        DFBCHECK(mLayer->GetCursorPosition(mLayer, &new_cx, &new_cy));
    }

    if (new_cx == mCursorX && new_cy == mCursorY)
        return;

    dx = new_cx - mCursorX;
    dy = new_cy - mCursorY;
    mCursorX = new_cx;
    mCursorY = new_cy;

    updateFocus();

    if (mPointerWindow) {
        we.type = DWET_MOTION;
        we.x = mCursorX - mPointerWindow->x();
        we.y = mCursorY - mPointerWindow->y();

        sendWindowEvent(mPointerWindow, we);
    } else if (mEnteredWindow){
        we.type = DWET_MOTION;
        we.x = mCursorX - mEnteredWindow->x();
        we.y = mCursorY - mEnteredWindow->y();

        sendWindowEvent(mEnteredWindow, we);
    }
}

void FL_DFB_WindowManager::do_queued_events()
{
    DFBEvent event;
    fd_set readset;
    struct timeval tv={0};//timeout is zero

    FD_ZERO(&readset);
    FD_SET(mEventFD, &readset);

    while(select(mEventFD + 1, &readset, NULL, NULL, &tv) > 0)
    { //check for data
        if (read(mEventFD, &event, sizeof(DFBEvent)) > 0)
        { //read an event
            switch(event.clazz) {
                case DFEC_INPUT:                    
                    //show_key_event((DFBInputEvent&)event);
                    processInput((DFBInputEvent&)event);
                    break;
                case DFEC_WINDOW:
                    {                    
                        sendWindowEventToFLTK((DFBWindowEvent&)event);
                        break;
                    }
                default:
                    D_TRACE("UNKNOW EVENT\n");
            }
        }
    }
}
#if ENABLE_IPC_INPUT
void FL_DFB_WindowManager::do_ipc_queued_events()
{
    DFBEvent event;
    SIpcMsgInf msg;
    int msgID;
    
    msgID = msgget(MSG_QUEUE_KEY, IPC_CREAT|0666);
    if (msgID < 0) {
        perror("Failed to get ipc input msg queue");
        return;
    }

    while(!mIPCExit) {
        ssize_t size;

        msg.m_type = 0;
        size = msgrcv(msgID, &msg, sizeof(msg) - sizeof(int), MSG_TYP_INPUTKEY_CMD, IPC_NOWAIT);
        if (size < 0 && errno != ENOMSG)
            perror("failed to rcv msg from msg queue");
        if (size > 0) {
            if (msg.m_type == MSG_TYP_INPUTKEY_CMD && msg.m_key_cod.clazz == DFEC_INPUT) {
                memcpy(&event, &msg.m_key_cod, sizeof(DFBInputEvent));
                //show_key_event((DFBInputEvent&)event);
                postEventToWM(&event);
                continue;
            } else {
                printf("got wrong msg from msg queue %d %d\n", msg.m_type, msg.m_key_cod.clazz);
            }
        }
        usleep(1000 * 10);
    }
    printf("%s: ipc thread exit %p\n", __FUNCTION__, pthread_self());
}
#endif

void FL_DFB_WindowManager::processInput(DFBInputEvent& event)
{
    if (event.flags & DIEF_BUTTONS)
        mButtons = event.buttons;
    if (event.flags & DIEF_MODIFIERS)
        mModifiers = event.modifiers;
    if (event.flags & DIEF_LOCKS)
        mLocks = event.locks;

    switch(event.type) {
        case DIET_KEYPRESS:
        case DIET_KEYRELEASE:
            {
                FL_DFB_Window* w = getKeyboardWindow();
                if (w) {
                    DFBWindowEvent we;

                    we.type = (event.type == DIET_KEYPRESS) ? DWET_KEYDOWN : DWET_KEYUP;
                    we.key_code = event.key_code;
                    we.key_id = event.key_id;
                    we.key_symbol = event.key_symbol;

                    sendWindowEvent(w, we);
                }
                break;
            }
        case DIET_BUTTONPRESS:
        case DIET_BUTTONRELEASE:
            {
                FL_DFB_Window* w = mPointerWindow ? mPointerWindow : mEnteredWindow;
                if (w) {
                    DFBWindowEvent we;

                    we.type = (event.type == DIET_BUTTONPRESS) ? DWET_BUTTONDOWN : DWET_BUTTONUP;
                    we.x = mCursorX - w->x();
                    we.y = mCursorY - w->y();
                    we.button = event.button;

                    sendWindowEvent(w, we);
                }
                break;
            }
        case DIET_AXISMOTION:
            {
                if (event.flags & DIEF_AXISREL) {
                    int rel = event.axisrel;
                    switch(event.axis) {
                        case DIAI_X:
                            processMotion(rel, 0);
                            break;
                        case DIAI_Y:
                            processMotion(0, rel);
                            break;
                        case DIAI_Z:
                            {
                                FL_DFB_Window* w = mPointerWindow ? mPointerWindow : mEnteredWindow;
                                if (w) {
                                    DFBWindowEvent we;

                                    we.type = DWET_WHEEL;
                                    we.x = mCursorX - w->x();
                                    we.y = mCursorY - w->y();
                                    we.step = -event.axisrel;
                                    
                                    sendWindowEvent(w, we);
                                }
                                break;
                            }
                        default:
                            ;
                    }
                } else if (event.flags & DIEF_AXISABS) {
                    if (event.axis == DIAI_X)
                        processMotion(event.axisabs - mCursorX, 0);
                    else if (event.axis == DIAI_Y)
                        processMotion(0, event.axisabs - mCursorY);
                }
                break;
            }
        default:
            ;
    }
}

