//
// "$Id: Fl_directfb.cxx,v 1.3 2006-10-31 09:31:14 nikego Exp $"
//
// DirectFB-specific code for the Fast Light Tool Kit (FLTK).
//
// Copyright 1998-2005 by Bill Spitzak and others.
// Written by Nikita Egorov (nikego@gmail.com)
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems on the following page:
//
//     http://www.fltk.org/str.php
//
// This file contains DirectFB-specific code for fltk which is always linked
// in.  Search other files for "DIRECTFB" or filenames ending in _directfb.cxx
// for other system-specific code.


#include <config.h>
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/x.H>
#include <sys/unistd.h>
#include <signal.h>
#  include <stdio.h>
#  include <stdlib.h>
#  include "flstring.h"
#  include <unistd.h>
#  include <sys/time.h>

#include "fl_directfb_eventfilter.h"

IDirectFB*              fl_dfb = 0; 
GC                      fl_screen = 0;
IDirectFBDisplayLayer*  fl_display_layer = 0;
GC                      fl_gc = 0;
DFBInputDeviceID        keyboard_id = 0;
DFBInputDeviceID        mouse_id = 0;
DFBEvent* fl_dfb_event = 0;

IDirectFBEventBuffer* fl_dfb_event_buffer = 0; // I have only one event buffer for all windows 
static int event_fd=-1;//
static Fl_DirectFB_EventFilter* fl_dfb_eventFilter = 0;
static FL_DFB_WindowManager* fl_dfb_wm;
static bool fl_exited = false;
static int fl_exit_code = 0;

u8 BG_COLOR_IDX=0;
u8 MASK_COLOR_IDX=1;

u8 BG_COLOR_RED=255,BG_COLOR_GREEN=0,BG_COLOR_BLUE=255;
u8 MASK_COLOR_RED=255,MASK_COLOR_GREEN=255,MASK_COLOR_BLUE=255; 

u32 BG_COLOR_RGB32 = PIXEL_RGB32(BG_COLOR_RED, BG_COLOR_GREEN, BG_COLOR_BLUE);
u32 MASK_COLOR_RGB32 = PIXEL_RGB32(MASK_COLOR_RED, MASK_COLOR_GREEN, MASK_COLOR_BLUE);

u16 BG_COLOR_RGB16 = PIXEL_RGB16(BG_COLOR_RED, BG_COLOR_GREEN, BG_COLOR_BLUE);
u16 MASK_COLOR_RGB16 = PIXEL_RGB16(MASK_COLOR_RED, MASK_COLOR_GREEN, MASK_COLOR_BLUE);

bool lut_ = false;
u32 pixf_ = DSPF_UNKNOWN;

Colormap fl_colormap;
VisualInfo visual;
FL_EXPORT VisualInfo *fl_visual = &visual;
ulong fl_event_time;

char fl_show_iconic;	// hack for iconize()

char fl_key_vector[32];

int screen_width=0, screen_height=0;

char last_dump_str[256];

Fl_Cursor current_cursor=FL_CURSOR_DEFAULT;
//extern Cursor current_xcursor;

void print_last_dump(){
  fprintf(stderr, "\ndumpstr:'%s'\n",last_dump_str);
}

ulong flip_size_sum = 0;

#define FLIP_FLAGS (DFBSurfaceFlipFlags)(/*DSFLIP_WAIT |*/ DSFLIP_BLIT)

// buffers flipping functions 
void fl_flip(GC gc, Region reg)
{
	if (gc){
        int W,H;
        get_size(gc, &W, &H);
		if (reg){
			flip_size_sum += (reg->x2-reg->x1+1)*(reg->y2-reg->y1+1);
		}else{
			flip_size_sum += W*H;
		}      
		flip_gc(gc,reg,FLIP_FLAGS);
	}
}

void post_id_event(DFBWindowID xid, DFBWindowEventType ev,int X, int Y,int cX=0,int cY=0)
{
	DFBEvent event;
	event.clazz = DFEC_WINDOW;
	event.window.type = ev;
	event.window.x = X;
	event.window.y = Y;
	event.window.cx = cX;
	event.window.cy = cY;
	event.window.window_id = xid;
	fl_dfb_event_buffer->PostEvent(fl_dfb_event_buffer, &event);
}

static int _show_cursor = 1;

void fl_show_cursor(int show)
{
  _show_cursor=show;
  if (fl_display_layer){
    _TRACE("cursor opacity %d\n", show);
    if (show){
      set_cursor_opacity(fl_display_layer,255);
    }else{
      set_cursor_opacity(fl_display_layer,0);
    }
  }
}

void post_win_event(Window win,DFBWindowEventType ev,int X, int Y){
	post_id_event(win, ev, X, Y);
}

void post_win_event(Fl_Window* fl_win,DFBWindowEventType ev,int X, int Y,int cX, int cY){
	post_id_event(Fl_X::i(fl_win)->xid, ev, X, Y, cX, cY);
}

static inline int can_boxcheat(uchar b) {return (b==1 || (b&2) && b<=15);}

void Fl_Window::show()
{
       set_visible();
  	image(Fl::scheme_bg_);
 	if (Fl::scheme_bg_){
    	labeltype(FL_NORMAL_LABEL);
    	align(FL_ALIGN_CENTER | FL_ALIGN_INSIDE | FL_ALIGN_CLIP);
  	}else{
    	labeltype(FL_NO_LABEL);
  	}
  	if (!shown()) {
  		fl_open_display();
    	if (can_boxcheat(box())) fl_background_pixel = int(fl_xpixel(color()));
    	Fl_X::make_xid(this);
    } else {
 		if (i->minimize()){
			i->delete_icon();
			redraw();
	  		i->first_draw=1;
		}
	}
}

Fl_Window* 
fl_find(dfbWindow _xid)
{
  Fl_X *window;
  for (Fl_X **pp = &Fl_X::first; (window = *pp); pp = &window->next) {
    if (window->_xid == _xid) {
      if (window != Fl_X::first && !Fl::modal()) {
	// make this window be first to speed up searches
	// this is not done if modal is true to avoid messing up modal stack
	*pp = window->next;
	window->next = Fl_X::first;
	Fl_X::first = window;
      }
      return window->w;
    }
  }
  return 0;
}

GC 
find_gc(XID xid)
{
  Fl_X *window;
  for (Fl_X **pp = &Fl_X::first;(window = *pp);pp = &window->next)
    if (window->xid == xid) 
      return window->gc;
  return 0;
}

Window fl_window;
Fl_Window *Fl_Window::current_;

void Fl_Window::make_current()
{
  fl_window  = i->xid;
  fl_gc      = i->gc;
  current_   = this;
  fl_clip_region(0);
}

// returns pointer to the filename, or null if name ends with '/'
const char *fl_filename_name(const char *name){
  const char *p,*q;
  if (!name) return (0);
  for (p=q=name; *p;) if (*p++ == '/') q = p;
  return q;
}

void Fl_Window::label(const char *name,const char *iname) {
	Fl_Widget::label(name);
	iconlabel_ = iname;
}

#  define CONSOLIDATE_MOTION 1
/**** Define this if your keyboard lacks a backspace key... ****/
/* #define BACKSPACE_HACK 1 */

////////////////////////////////////////////////////////////////
// interface to poll/select call:

#  if USE_POLL
#    include <poll.h>
static pollfd *pollfds = 0;
#  else
#    if HAVE_SYS_SELECT_H
#      include <sys/select.h>
#    endif /* HAVE_SYS_SELECT_H */

// The following #define is only needed for HP-UX 9.x and earlier:
//#define select(a,b,c,d,e) select((a),(int *)(b),(int *)(c),(int *)(d),(e))

static fd_set fdsets[3];
static int maxfd;
#    define POLLIN 1
#    define POLLOUT 4
#    define POLLERR 8

#  endif /* USE_POLL */

static int nfds = 0;
static int fd_array_size = 0;
struct FD {
#  if !USE_POLL
  int fd;
  short events;
#  endif
  void (*cb)(int, void*);
  void* arg;
};

static FD *fd = 0;

void Fl::add_fd(int n, int events, void (*cb)(int, void*), void *v) {
  remove_fd(n,events);
  int i = nfds++;
  if (i >= fd_array_size) {
    FD *temp;
    fd_array_size = 2*fd_array_size+1;

    if (!fd) temp = (FD*)malloc(fd_array_size*sizeof(FD));
    else temp = (FD*)realloc(fd, fd_array_size*sizeof(FD));

    if (!temp) return;
    fd = temp;
 
#  if USE_POLL
    pollfd *tpoll;

    if (!pollfds) tpoll = (pollfd*)malloc(fd_array_size*sizeof(pollfd));
    else tpoll = (pollfd*)realloc(pollfds, fd_array_size*sizeof(pollfd));

    if (!tpoll) return;
    pollfds = tpoll;
#  endif
  }
  fd[i].cb = cb;
  fd[i].arg = v;
#  if USE_POLL
  pollfds[i].fd = n;
  pollfds[i].events = events;
#  else
  fd[i].fd = n;
  fd[i].events = events;
  if (events & POLLIN) FD_SET(n, &fdsets[0]);
  if (events & POLLOUT) FD_SET(n, &fdsets[1]);
  if (events & POLLERR) FD_SET(n, &fdsets[2]);
  if (n > maxfd) maxfd = n;
#  endif
}

void Fl::add_fd(int n, void (*cb)(int, void*), void* v) {
  Fl::add_fd(n, POLLIN, cb, v);
}

void Fl::remove_fd(int n, int events) {
  int i,j;
  maxfd = -1; // recalculate maxfd on the fly
  for (i=j=0; i<nfds; i++) {
#  if USE_POLL
    if (pollfds[i].fd == n) {
      int e = pollfds[i].events & ~events;
      if (!e) continue; // if no events left, delete this fd
      pollfds[j].events = e;
    }
#  else
    if (fd[i].fd == n) {
      int e = fd[i].events & ~events;
      if (!e) continue; // if no events left, delete this fd
      fd[i].events = e;
    }
#  endif
    if (fd[i].fd > maxfd) maxfd = fd[i].fd;
    // move it down in the array if necessary:
    if (j<i) {
      fd[j] = fd[i];
#  if USE_POLL
      pollfds[j] = pollfds[i];
#  endif
    }
    j++;
  }
  nfds = j;
#  if !USE_POLL
  if (events & POLLIN) FD_CLR(n, &fdsets[0]);
  if (events & POLLOUT) FD_CLR(n, &fdsets[1]);
  if (events & POLLERR) FD_CLR(n, &fdsets[2]);
#  endif
}

void Fl::remove_fd(int n) {
  remove_fd(n, -1);
}
int grabbed=0;
int startx,endx;
int starty,endy;
int new_W=0,new_H=0;
FL_DFB_Window *grabbed_win=0;
int on_border = 0;
Fl_Window *resized_win=0;
Fl_Window* resize_bug_fix = 0;

#if CONSOLIDATE_MOTION
static Fl_Window* send_motion;
extern Fl_Window* fl_xmousewin;
#endif
static bool in_a_window; // true if in any of our windows, even destroyed ones
static DFBEvent oldWindowEvent;
static int gnashReady = 1;

static void do_queued_events() {
  	in_a_window = true;
    DFBEvent event;
    fd_set readset;
    FD_ZERO(&readset);
    FD_SET(event_fd, &readset);
    struct timeval tv={0};//timeout is zero
    while(select(event_fd+1,&readset,NULL,NULL,&tv)>0){ //check for data
        if (read(event_fd,&event,sizeof(DFBEvent))>0)  { //read an event

            if (fl_dfb_eventFilter && !fl_dfb_eventFilter->eventFilter(event))
			    continue;
			
            fl_dfb_event = &event;
			
	        if (fl_handle((const DFBWindowEvent&)event) == 2) {    //handle the event	        
	        	oldWindowEvent = event;
	        	fl_dfb_event = 0;
	        	gnashReady = 0;
	        	
				DFBEvent null_event = event;
				null_event.window.key_symbol = DIKS_NULL;
				null_event.window.key_id = DIKI_UNKNOWN;
				fl_dfb_event_buffer->PostEvent(fl_dfb_event_buffer, &null_event);	        	
	        	break;
	        }
	        	
	        	
            fl_dfb_event = 0;
        }
    }

  // we send FL_LEAVE only if the mouse did not enter some other window:
  	if (!in_a_window) Fl::handle(FL_LEAVE, 0);
#if CONSOLIDATE_MOTION
  	else if (send_motion == fl_xmousewin) {
    	send_motion = 0;
    	Fl::handle(FL_MOVE, fl_xmousewin);
  	}
#endif
 	if (grabbed&&(grabbed_win||resized_win)){
 		if (startx != endx || starty != endy){
 			if (!on_border){
 				if (grabbed_win)
					grabbed_win->move(endx-startx,endy-starty);
				startx = endx;
        		starty = endy;
 			}else 
 			  if (resized_win) {
 				int W = resized_win->w()+endx-startx;
 				int H = resized_win->h()+endy-starty;
 				Fl_X* xw=Fl_X::i(resized_win);
 				int limitw=0,limith=0; 
 				if (xw->minw && W<xw->minw){W=xw->minw;limitw=1;}
 				if (xw->minh && H<xw->minh){H=xw->minh;limith=1;}
 				if (xw->maxw && W>xw->maxw){W=xw->maxw;limitw=1;}
 				if (xw->maxh && H>xw->maxh){H=xw->maxh;limith=1;} 
        		resized_win->size(W, H);
        		if (!limitw)startx=endx;
 				if (!limith)starty=endy;
 			}
 		}
  	}
}

// these pointers are set by the Fl::lock() function:
static void nothing() {}
void (*fl_lock_function)() = nothing;
void (*fl_unlock_function)() = nothing;

// This is never called with time_to_wait < 0.0:
// It should return negative on error, 0 if nothing happens before
// timeout, and >0 if any callbacks were done.
int fl_wait(double time_to_wait) {

#  if !USE_POLL
  fd_set fdt[3];
  fdt[0] = fdsets[0];
  fdt[1] = fdsets[1];
  fdt[2] = fdsets[2];
#  endif
  int n;

  fl_unlock_function();

  if (time_to_wait < 2147483.648) {
#  if USE_POLL
    n = ::poll(pollfds, nfds, int(time_to_wait*1000 + .5));
#  else
    timeval t;
    t.tv_sec = int(time_to_wait);
    t.tv_usec = int(1000000 * (time_to_wait-t.tv_sec));
    n = ::select(maxfd+1,&fdt[0],&fdt[1],&fdt[2],&t);
#  endif
  } else {
#  if USE_POLL
    n = ::poll(pollfds, nfds, -1);
#  else
    n = ::select(maxfd+1,&fdt[0],&fdt[1],&fdt[2],0);
#  endif
  }

  fl_lock_function();

  if (n > 0) {
    for (int i=0; i<nfds; i++) {
#  if USE_POLL
      if (pollfds[i].revents) fd[i].cb(pollfds[i].fd, fd[i].arg);
#  else
      int f = fd[i].fd;
      short revents = 0;
      if (FD_ISSET(f,&fdt[0])) revents |= POLLIN;
      if (FD_ISSET(f,&fdt[1])) revents |= POLLOUT;
      if (FD_ISSET(f,&fdt[2])) revents |= POLLERR;
      if (fd[i].events & revents) fd[i].cb(f, fd[i].arg);
#  endif
    }
  }
  return n;
}

// fl_ready() is just like fl_wait(0.0) except no callbacks are done:
int fl_ready() {
#  if USE_POLL
  return ::poll(pollfds, nfds, 0);
#  else
  timeval t;
  t.tv_sec = 0;
  t.tv_usec = 0;
  fd_set fdt[3];
  fdt[0] = fdsets[0];
  fdt[1] = fdsets[1];
  fdt[2] = fdsets[2];
  return ::select(maxfd+1,&fdt[0],&fdt[1],&fdt[2],&t);
#  endif
}

static void fd_callback(int dfd,void *) {
	if (gnashReady)	
		do_queued_events(); 
    else {
		fl_dfb_event = &oldWindowEvent;
		
		if (fl_handle((const DFBWindowEvent&)oldWindowEvent) == 2) {    //handle the event
			/*DFBEvent null_event = oldWindowEvent;
			null_event.window.key_symbol = DIKS_NULL;
			null_event.window.key_id = DIKI_UNKNOWN;
			fl_dfb_event_buffer->PostEvent(fl_dfb_event_buffer, &null_event);		
			*/
			usleep(0);//let gnash process has chance to initialize.		
		} else
			gnashReady = 1;	        	
		fl_dfb_event = 0;
  	}
}
extern int id2fltk(int id);

static int px, py;
static ulong ptime;

static void set_event_xy(DFBWindowEvent& xevent) {
#  if CONSOLIDATE_MOTION
  	send_motion = 0;
#  endif
  	Fl::e_x_root  = xevent.cx;
  	Fl::e_x       = xevent.x;
  	Fl::e_y_root  = xevent.cy;
  	Fl::e_y       = xevent.y;
  	//Fl::e_state   = xevent.buttons << 24;
  	fl_event_time = xevent.timestamp.tv_sec*1000 + xevent.timestamp.tv_usec / 1000;
#  ifdef __sgi
  // get the meta key off PC keyboards:
  	if (fl_key_vector[18]&0x18) Fl::e_state |= FL_META;
#  endif
	if (xevent.button == DIBI_RIGHT){
		xevent.button = DIBI_MIDDLE;
		if (xevent.type == DWET_BUTTONDOWN){ 
			xevent.buttons =(DFBInputDeviceButtonMask)(((int)xevent.buttons & ~((int)DIBM_RIGHT))|((int)DIBM_MIDDLE));
		}
	}else if (xevent.button ==DIBI_MIDDLE){
		xevent.button = DIBI_RIGHT;
		if (xevent.type == DWET_BUTTONDOWN){ 
			xevent.buttons =(DFBInputDeviceButtonMask)(((int)xevent.buttons & ~((int)DIBM_MIDDLE)) | ((int)DIBM_RIGHT));
		}
	}
// turn off is_click if enough time or mouse movement has passed: DIBI_LEFT 
  	if (abs(Fl::e_x_root-px)+abs(Fl::e_y_root-py) > 3 ||
      	fl_event_time >= ptime+1000)
    Fl::e_is_click = 0;
}

const int WHEADER_H = 30;//height of window's header
const int WBORDER_W = 2;//width of border
const int WBORDER_H = 2;//height of border

//the function returns state of cursor position
// state is one of followed values:
// 0 - out of border
// 1 - vertical border
// 2 - horizontal border
// 3 - both (right-bottom corner)
int is_on_border(Fl_Window* window, int x, int y)
{
	dfbWindow w = Fl_X::i(window)->_xid;
	int W,H,X,Y,ret = 0;
	get_size(w,&W,&H);
	get_position(w,&X,&Y);
	if ((x>=X+W-WBORDER_W-1)&&(x<=X+W)&&(y<=Y+H)&&(y>=Y)){
		ret |= 1;
		if (y>=Y+H-10)
			ret|= 2;
	}
	if ((y>=Y+H-WBORDER_H-1)&&(y<=Y+H)&&(x<=X+W)&&(x>X)){
		ret |= 2;
		if (x>=X+W-10)
			ret|=1;
	}
	return ret;	
}

Fl_Window* check_event_pos(DFBWindowEvent& xevent){
	Fl_Window* window = fl_find(xevent.window_id);
	Fl_Window* header = 0;
    //static Cursor saved_cursor = 0;
	if (!window)
		return 0;
	if (window->border()){
		int win_x = window->x(),maximize=Fl_X::i(window)->maximize();
		header = Fl_X::i(window)->header;
		//check for window border (need resize ?)
		if (window->resizable()){
			if (header && !header->child(2)->active())
				header->child(2)->activate();
	    	int b = is_on_border(window,xevent.cx,xevent.cy); 
	    	if ((b!=on_border)&&!grabbed&&!(Fl::e_state&(0xF<<24))){
	    		if (b){
	    			if (on_border!=0)//free cursor if previous state is "on border"
                       XDefineCursor(fl_display, Fl_X::i(window)->xid, 0);
	    			//	saved_cursor = current_xcursor;
	    			switch (b){
	    				case 1:	fl_cursor(FL_CURSOR_WE);break;
	    				case 2:	fl_cursor(FL_CURSOR_NS);break;
	    				case 3:	fl_cursor(FL_CURSOR_NWSE);break;
	    			}
	    			grabbed_win=Fl_X::i(window)->_xid;
		           	grabbed_win->grabPointer();
		        }else{
                    XDefineCursor(fl_display, Fl_X::i(window)->xid/*xevent.window_id*/, 0);
	    			//XDefineCursor(fl_display, Fl_X::i(window)->xid/*xevent.window_id*/, saved_cursor);
	    			if (grabbed_win){	
	    				grabbed_win->ungrabPointer();
		   				grabbed_win=0;
	    			}
	    			grabbed=0;
		   		}
		   		on_border = b;
	    	}
    	}else{
    		if (header && header->child(2)->active())
				header->child(2)->deactivate();
		}
    	//start resizing if user press left mouse button on the border
		if (on_border&&(!maximize)&&!grabbed&&(xevent.type==DWET_BUTTONDOWN)&&(xevent.button==DIBI_LEFT)){
	    	resized_win=window;
			grabbed=1;
			endx=startx=xevent.cx;
	        endy=starty=xevent.cy;
	    }
	   	if ((xevent.y > 0) && (xevent.y < WHEADER_H) && !(Fl::e_state&(0xF<<24))){//check for header's place
    		if (header){ //if header is present
    			if  (Fl::pushed()){Fl::handle(FL_RELEASE, window);}
    			xevent.x = xevent.cx+WBORDER_W;// restore correct xy value for header
    			xevent.y = xevent.cy+WHEADER_H;
    			window = header;// change the current window to the header
    		//start to drag the window when user press left button on the header
    			if ((!maximize)&&(xevent.type==DWET_BUTTONDOWN)){
    	   			if (xevent.x<win_x+header->w()-((WHEADER_H-5)+2)*3){//avoid buttons place
		   		    	if (!grabbed && (xevent.button==DIBI_LEFT)){
		 					endx=startx=xevent.cx;
		            		endy=starty=xevent.cy;
		            		grabbed_win = Fl_X::i(window)->_xid;
		            		grabbed_win->grabPointer();
		            		grabbed=1;
		    			}
		    		}
		   		}
	   		}
    	}else{
    		xevent.x -= WBORDER_W; //shift the position to subsurface 
	    	xevent.y -= WHEADER_H;
	   	}
	    //exit from dragging/resizing mode
	    if (xevent.type==DWET_BUTTONUP){
	       	if (grabbed && (xevent.button==DIBI_LEFT)){
	       		if (resized_win){
                    //XDefineCursor(fl_display, Fl_X::i(window)->xid, 0);
	    			//XDefineCursor(fl_display, Fl_X::i(window)->xid/*xevent.window_id*/, saved_cursor);
	    			resized_win=0;
	    		}
	    		grabbed=0;
	    		goto release_grabbed_win;
	    	}
	    }
    }
    //FIXME if user press button then I grab mouse to avoid missing of event "ButtonUp"
    if ((header!=window) && !grabbed_win && (xevent.type==DWET_BUTTONDOWN)){
    	grabbed_win = Fl_X::i(window)->_xid;
    	grabbed_win->grabPointer();
     }
    //resp. I ungrab mouse 
    if ((header!=window)&&(xevent.type==DWET_BUTTONUP)){
release_grabbed_win:    	
		if (grabbed_win){
	   		grabbed_win->ungrabPointer();
	   		grabbed_win=0;
	    }
    }
    return window;
}

// if this is same event as last && is_click, increment click count:
inline void checkdouble() {
  	if (Fl::e_is_click == Fl::e_keysym)
    	Fl::e_clicks++;
  	else {
    	Fl::e_clicks = 0;
    	Fl::e_is_click = Fl::e_keysym;
  	}
  	px = Fl::e_x_root;
  	py = Fl::e_y_root;
  	ptime = fl_event_time;
}

inline int check_modif (int& state, DFBWindowEvent& xevent){
	state = 0;
	if (xevent.modifiers & DIMM_SHIFT) state |= FL_SHIFT;
   	if (xevent.modifiers & DIMM_ALT) state |= FL_ALT;
   	if (xevent.modifiers & DIMM_CONTROL) state |= FL_CTRL;
   	if (xevent.modifiers & DIMM_META) state |= FL_META;
   	if (xevent.locks & DILS_SCROLL) state |= FL_SCROLL_LOCK;
   	if (xevent.locks & DILS_NUM) state |= FL_NUM_LOCK;	
   	if (xevent.locks & DILS_CAPS) state |= FL_CAPS_LOCK;
   	state |= xevent.buttons << 24;
   	return state;
}
#define WHEEL_DELTA 1
int ts = 0;

int fl_handle(const DFBWindowEvent& thisevent){
	DFBWindowEvent xevent = thisevent;
  	Fl_Window *window = fl_find(xevent.window_id);
    int ret=0;
 	if (!window) return 0;
   	switch (xevent.type) {
  	case DWET_BUTTONDOWN:
  	  	window=check_event_pos(xevent);
  		set_event_xy(xevent);
  		Fl::e_keysym = FL_Button + xevent.button + 1;
  		check_modif (Fl::e_state, xevent);	
    	in_a_window = true;
       	checkdouble();
        if (!Fl::handle(FL_PUSH, window)) return 0;
    	return 1;
  	case DWET_BUTTONUP:
  		window=check_event_pos(xevent);
  		set_event_xy(xevent);
  		Fl::e_keysym = FL_Button + xevent.button + 1;
   		//Fl::e_state &= ~(FL_BUTTON1 << xevent.button);
    	check_modif (Fl::e_state, xevent);	
  		in_a_window = true;
    	if (!Fl::handle(FL_RELEASE, window)) return 0;
  		return 1;
	case DWET_MOTION:
		window=check_event_pos(xevent);
		set_event_xy(xevent);
		in_a_window = true;
		if (grabbed){
			if ((on_border == 0)||(on_border&1))
		    	endx = xevent.cx;
		    if ((on_border == 0)||(on_border&2))	
            	endy = xevent.cy;
  		}
		//Fl::e_state &= ~(FL_BUTTON1 << xevent.button);
    	//check_modif (Fl::e_state,xevent);	
    	return Fl::handle(FL_MOVE, window);
	case DWET_WHEEL:
		window=check_event_pos(xevent);
		set_event_xy(xevent);
		static int delta = 0; // running total of all motion
    	delta += xevent.step;
    	Fl::e_dy = -delta / WHEEL_DELTA;
    	delta += Fl::e_dy * WHEEL_DELTA;
		return Fl::handle(FL_MOUSEWHEEL, window);
	case DWET_KEYUP:
	case DWET_KEYDOWN:{
		fl_event_time = xevent.timestamp.tv_sec*1000 + xevent.timestamp.tv_usec / 1000;
//		static int pressed = 0;
//		if ((pressed==thisevent.key_id)&&(xevent.type==DWET_KEYDOWN))
//			return 0;
		int fltk = id2fltk(thisevent.key_id);
		if (!fltk)
			fltk = thisevent.key_symbol;
    	Fl::e_keysym = fltk;
		static char buffer[2];
		if (thisevent.key_symbol < 256){
			buffer[0] = thisevent.key_symbol;
			Fl::e_length = 1;
		}else{
			buffer[0] = 0;
			Fl::e_length = 0;
		}
		check_modif (Fl::e_state,xevent);
		if (Fl::e_state & FL_CTRL){ //FIXME I change symbol if user pressed Ctrl
			if (Fl::e_length){
				switch (buffer[0]){
				case 13: buffer[0]=10;break;
				case '[':case '\\':case ']':buffer[0]-= 'A'-1;break;
				case 8:	buffer[0]=127;break;
				default:
					if (buffer[0]>='a' && buffer[0] <='z')
						buffer[0]-='a'-1;
					else{
						buffer[0] = 0;
						Fl::e_length = 0;
					}	
				}
			}
		}
		Fl::e_text = buffer;
		// FIXME if I set e_is_click to NULL then adjuster is not working during key is held
		if (xevent.type == DWET_KEYUP)
	   		Fl::e_is_click = 0;
	   	int keycode = thisevent.key_code;
	   	in_a_window = true;
	   	int key_handle_ret = 1;
		if (xevent.type == DWET_KEYDOWN){
//			if (DIKI_SHIFT_L <= thisevent.key_id && DIKI_HYPER_R >= thisevent.key_id){
//				pressed = thisevent.key_id;
//			}
 	    	fl_key_vector[keycode/8] |= (1 << (keycode%8));
	    	key_handle_ret = Fl::handle(FL_KEYBOARD, window);
	    	if (key_handle_ret == 0)
	    	{
				if(fl_dfb_event->window.key_symbol == DIKS_TITLE)
				{
					key_handle_ret = Fl::handle(FL_KEYBOARD, window);
					//printf("@@@@@@ %s:%d %s() return %d.\n", __FILE__, __LINE__, __FUNCTION__, key_handle_ret);
				}
	    	}
	    	//printf("@@@@@@ %s:%d %s() return %d.\n", __FILE__, __LINE__, __FUNCTION__, key_handle_ret);
    	}else{
//    		if (pressed == thisevent.key_id)
//    			pressed = 0;
    		fl_key_vector[keycode/8] &= ~(1 << (keycode%8));
    	   	key_handle_ret = Fl::handle(FL_KEYUP, window);
    	   	//printf("@@@@@@ %s:%d %s() return %d.\n", __FILE__, __LINE__, __FUNCTION__, key_handle_ret);
    	}
  		return key_handle_ret;
	}
	case DWET_POSITION_SIZE: //window created
		if (!window->border()){
			Fl_X::x(window,xevent.x);Fl_X::y(window,xevent.y);
		}else{
			Fl_X::x(window,xevent.x+WBORDER_W);Fl_X::y(window,xevent.y+WHEADER_H);			
		}
	  	Fl_X::i(window)->wait_for_expose = 0;
		return Fl::handle(FL_SHOW, window);
	case DWET_POSITION:
		if (!window->border()){
			Fl_X::x(window,xevent.x);Fl_X::y(window,xevent.y);
		}else{
			Fl_X::x(window,xevent.x+WBORDER_W);Fl_X::y(window,xevent.y+WHEADER_H);			
		}
		return 1;
	case DWET_SIZE:
	{	int W = xevent.w, H = xevent.h;
		if (window->border() && !ts){ W-=WBORDER_W*2;	H-=WHEADER_H+WBORDER_H; }
		resize_bug_fix = window;
		window->resize(window->x(), window->y(), W, H);
        _TRACE("DWET_SIZE \n");
		//Fl_X::i(window)->set_surface();
    	Fl_X::i(window)->first_draw = 1;
		Fl_X::i(window)->wait_for_expose = 0;
		return 1;
	}
	case DWET_CLOSE:
		TRACE("handle event 'DWET_CLOSE'\n");
	case DWET_DESTROYED:
		TRACE("handle event 'DWET_DESTROYED'\n");
		return Fl::handle(FL_CLOSE, window);
		break;
	case DWET_GOTFOCUS:
		ret = Fl::handle(FL_FOCUS, window);
		return ret;
	case DWET_LOSTFOCUS:
		ret = Fl::handle(FL_UNFOCUS, window);
		return ret;
	case DWET_ENTER:
		check_event_pos(xevent);
		ret = Fl::handle(FL_ENTER,window);
		Fl_X::i(window)->enter_window(true);
		return ret;
	case DWET_LEAVE:
		check_event_pos(xevent);
		Fl::handle(FL_LEAVE, window);
		Fl_X::i(window)->enter_window(false);
		return ret;
	default:
		_TRACE("handle event 'UNKNOWN' %d\n", xevent.type);
 	}
  	return 0;
}
  		
void Fl_Window::resize(int X,int Y,int W,int H)
{
  int is_a_move = (X != x() || Y != y());
  int is_a_resize = (W != w() || H != h());

  int resize_from_program = (this != resize_bug_fix);
  if (!resize_from_program) resize_bug_fix = 0;

  if (shown() && i->maximize()){
    return;
  }
  if (parent()){//it is not a real window - we should change surface position only
    if (X < 0) X = 0;
    if (Y < 0) Y = 0;
    if (H <=0) H = 1;
    if (W <=0) W = 1;
    Fl_Group::resize(X,Y,W,H);
    if (shown()){
      redraw();
      Fl_X *parent_xid = get_parent_xid(this);
      int max_x = parent()->x()+parent()->w()-1;
      int max_y = parent()->y()+parent()->h()-1;
      x(X>=max_x?max_x-1:X);
      y(Y>=max_y?max_y-1:Y);
      w(W>0 ? W : 1); h(H>0 ? H : 1);
      release_gc(i->gc);
      //create subsurface
      DFBRectangle rect = {x(),y(),w(),h()};
      if (parent_xid->header == this){//it's a header ?
        i->gc = create_gc(parent_xid->win_gc, &rect);
      }else{
        i->gc = create_gc(parent_xid->gc, &rect);
      }
    }
    return;
  }

  if (is_a_move && resize_from_program)
    set_flag(FL_FORCE_POSITION);
  else if (!is_a_resize && !is_a_move)
    return;
  if (is_a_resize){
    Fl_Group::resize(X,Y,W,H);
    if (shown()){
      redraw();
      i->wait_for_expose = 1;
    }
  }else{
    x(X); y(Y);
  }

  if (resize_from_program && is_a_resize && !resizable()){
    size_range(w(), h(), w(), h());
  }

  if (resize_from_program && shown()){
    if (is_a_resize){
      if (!resizable()) size_range(w(),h(),w(),h());
      if (is_a_move){
        if (border()){X-=WBORDER_W;Y-=WHEADER_H;}
        i->_xid->moveTo(X, Y);
      }
      if (border()){W+=WBORDER_W*2; H+=WHEADER_H+WBORDER_H;}
      if (i->_xid)
        i->_xid->resize(W>0 ? W : 1, H>0 ? H : 1);
      i->set_gc();
    }else{
      if (border()){X-=WBORDER_W;Y-=WHEADER_H;}
      i->_xid->moveTo(X, Y);
    }
  }
}
////////////////////////////////////////////////////////////////

Fl_Widget *fl_selection_requestor;
char *fl_selection_buffer[2];
int fl_selection_length[2];
int fl_selection_buffer_length[2];
char fl_i_own_selection[2];

int Fl::x(){
  return 0;
}

int Fl::y(){
  return 0;
}

int Fl::h(){
  if (!screen_height)
    fl_open_display();
  return screen_height;
}

int Fl::w(){
  if (!screen_width)
    fl_open_display();
  return screen_width;
}

void Fl::get_mouse(int &xx, int &yy){
  fl_open_display();
  get_cursor_position(fl_display_layer, &xx, &yy);
}

void Fl_Window::size_range_() {
  size_range_set = 1;
  if (shown()) i->sendxjunk();
}

void Fl_X::sendxjunk(){
  if (w->parent() || w->override()) return; // it's not a window manager window!

  if (!w->size_range_set) { // default size_range based on resizable():
    if (w->resizable()) {
      Fl_Widget *o = w->resizable();
      int minw = o->w(); if (minw > 100) minw = 100;
      int minh = o->h(); if (minh > 100) minh = 100;
      w->size_range(w->w() - o->w() + minw, w->h() - o->h() + minh, 0, 0);
    } else {
      w->size_range(w->w(), w->h(), w->w(), w->h());
    }
    return; // because this recursively called here
  }
  	minw=w->minw;minh=w->minh;maxw=w->maxw;maxh=w->maxh;
	dw=w->dw;dh=w->dh; 
  //the handle of  the events "border <-> no border" 
	if ((bool)(w->border()) != (bool)(header)){
		int X=w->x(),Y=w->y(),W=w->w(),H=w->h();
		get_position(_xid, &X, &Y);
		get_size(_xid, &W, &H);
		int dx,dy;
		if (!w->border()){
			//get_position(xid, &X, &Y);
			//get_size(xid, &W, &H);
			H-=WHEADER_H+WBORDER_H;
			W-=WBORDER_W*2;
			dx=WBORDER_W;
			dy=WHEADER_H;
		}else{
			dx=-WBORDER_W;
			dy=-WHEADER_H;
			H+=WHEADER_H+WBORDER_H;
			W+=WBORDER_W*2;
		}
		ts = 0;
		_xid->resize(W,H);
		_xid->move(dx,dy); 
		ts = 0;
		w->Fl_Group::resize(X,Y,W,H);//resize of group (!!!)
		DFBEvent event;
		event.clazz = DFEC_WINDOW;
		event.window.type = DWET_POSITION;
		event.window.x = X+dx; event.window.y = Y+dy;
		event.window.window_id = xid;
		fl_dfb_event_buffer->PostEvent(fl_dfb_event_buffer, &event);
		event.window.type = DWET_ENTER;
		Fl::get_mouse(event.window.cx, event.window.cy);
		fl_dfb_event_buffer->PostEvent(fl_dfb_event_buffer, &event);
		set_gc();
		first_draw = 1;
		w->redraw();
	}
}

// Call this when a "paste" operation happens:
void Fl::paste(Fl_Widget &receiver, int clipboard) {
  if (fl_i_own_selection[clipboard]) {
    // We already have it, do it quickly without window server.
    // Notice that the text is clobbered if set_selection is
    // called in response to FL_PASTE!
    Fl::e_text = fl_selection_buffer[clipboard];
    Fl::e_length = fl_selection_length[clipboard];
    if (!Fl::e_text) Fl::e_text = (char *)"";
    receiver.handle(FL_PASTE);
    return;
  }
  // otherwise get the window server to return it:
  fl_selection_requestor = &receiver;
}

////////////////////////////////////////////////////////////////
// Code for copying to clipboard and DnD out of the program:

void Fl::copy(const char *stuff, int len, int clipboard) {
  if (!stuff || len<0) return;
  if (len+1 > fl_selection_buffer_length[clipboard]) {
    delete[] fl_selection_buffer[clipboard];
    fl_selection_buffer[clipboard] = new char[len+100];
    fl_selection_buffer_length[clipboard] = len+100;
  }
  memcpy(fl_selection_buffer[clipboard], stuff, len);
  fl_selection_buffer[clipboard][len] = 0; // needed for direct paste
  fl_selection_length[clipboard] = len;
  fl_i_own_selection[clipboard] = 1;
  //NIKITA Atom property = clipboard ? CLIPBOARD : XA_PRIMARY;
  //NIKITA XSetSelectionOwner(fl_display, property, fl_message_window, fl_event_time);
}

extern void fl_fix_focus(); 

Fl_X* Fl_X::set_xid(Fl_Window* window, dfbWindow win)
{
  Fl_X* xp        = new Fl_X;
  xp->gc          = xp->win_gc = 0;
  xp->header      = 0;
  xp->setwindow(window);

  xp->next            = Fl_X::first;
  Fl_X::first         = xp;
  xp->region          = 0;
  xp->first_draw      = 1;
  xp->wait_for_expose = 0;
  xp->backbuffer_bad  = 1;
  xp->origin_size.w   = 0;
  xp->cursors[0]      = 0;
  
  //DFBWindowID win_id  = -1;
  //win->GetID(win, &win_id);
  

  xp->xid       = win->getXID();
  xp->_xid      = win;
   __TRACE("set_xid() id = %d\n",xp->xid);
  xp->minimize_ = 0;
  if (window->modal()){
    Fl::modal_ = window;
    fl_fix_focus();
  }
  return xp;
}

int fl_background_pixel = -1; // hack to speed up bg box drawing
int fl_disable_transient_for; // secret method of removing TRANSIENT_FOR

extern void init_colormap();

#include <FL/Fl_Button.H>
#include <FL/Fl_Box.H>

Fl_X *get_parent_xid(Fl_Widget* w)
{
  Fl_Widget* parent_wdg = w->parent();
  Fl_Window* parent_win = (Fl_Window*)parent_wdg;
  if (parent_wdg->type() < FL_WINDOW){
    parent_win = parent_wdg->window();
  }
  return Fl_X::i(parent_win);
}

/*****************************************************************
 * 
 * header and icon creating and handling
 * 
 */
 
#include <FL/Fl_Bitmap.H>
#include <FL/Fl_Pixmap.H>

extern unsigned char idata_close_but[];
extern unsigned char idata_max_but[];
extern unsigned char idata_min_but[];
extern unsigned char idata_norm_but[];
extern const char* idata_dfb[];

static Fl_Bitmap image_close_but(idata_close_but, 12, 10);
static Fl_Bitmap image_max_but(idata_max_but, 12, 10);
static Fl_Bitmap image_min_but(idata_min_but, 12, 10);
static Fl_Bitmap image_norm_but(idata_norm_but, 12, 10);
static Fl_Pixmap image_icon(idata_dfb);

Fl_Window* create_icon_win(int& W, int H, Fl_Window*);

int current_icon_pos = 0;

void on_close_btn(Fl_Widget* pThis, void *param){
  if (param){
    Fl_Window* win = (Fl_Window*)param;
    Fl::handle(FL_CLOSE,win);
  }
}
void on_icon_btn(Fl_Widget* pThis, void *param){
  if (param){
    Fl_Window* win = (Fl_Window*)param;
    if (Fl::modal() != win){
      win->iconize();
    }
  }
}
void on_full_btn(Fl_Widget* pThis, void *param){
  if (param){
    Fl_Window* win = (Fl_Window*)param;
    if (!win->resizable())
      return;
    Fl_X *xid = Fl_X::i(win);
    if (!xid->maximize()){
      xid->set_maximize();
      pThis->image(image_norm_but);
    }else{
      xid->restore_size();
      pThis->image(image_max_but);
    }
  }
}
#define HEADER_LABEL "__header__"
#define ICON_LABEL   "__icon__"

Fl_Window* create_header(int W, int H, Fl_Window* win)
{
	Fl_Window* header = new Fl_Window(0,0,W,H,HEADER_LABEL);
	header->border(0);
	header->box(FL_THIN_UP_BOX);
	header->color(Fl_Color(12));
	header->selection_color(FL_DARK_BLUE);
	int but_h = H-6 ,but_w= H-5 ;	
	int pos_x = W-(but_w+2)*3-1;  
	Fl_Box* box = new Fl_Box(5,0,pos_x-5,H);
	if (win->iconlabel())
		box->label(win->iconlabel());
	else 
		if (win->label())	
			box->label(win->label());
		else
			box->label("_");
    Fl_Boxtype box_type = /*(!have_lut()) ? FL_PLASTIC_UP_BOX :*/ FL_THIN_UP_BOX;
	box->labelcolor(FL_WHITE);
	box->labelfont(FL_COURIER_BOLD);
	box->align(FL_ALIGN_LEFT | FL_ALIGN_INSIDE);
    box->labeltype(FL_NORMAL_LABEL);
	header->resizable(box);
    //box = new Fl_Box(5,0,pos_x-5,H);
    //box->image(image_icon);
	Fl_Button *btn = new Fl_Button(pos_x,3,but_w,but_h);
	btn->image(image_min_but);
	btn->box(box_type);
	btn->labeltype(FL_NORMAL_LABEL);
	//btn->align(FL_ALIGN_TOP | FL_ALIGN_INSIDE);
	btn->callback(on_icon_btn);
	btn->when(FL_WHEN_RELEASE_ALWAYS);
	btn->user_data((void*)win);
	btn = new Fl_Button(pos_x+but_w+2,3,but_w,but_h);
	btn->image(image_max_but);
	btn->box(box_type);
	btn->labeltype(FL_NORMAL_LABEL);
	//btn->align(FL_ALIGN_TOP | FL_ALIGN_INSIDE);
	btn->callback(on_full_btn);
	btn->when(FL_WHEN_RELEASE_ALWAYS);
	btn->user_data((void*)win);
	btn = new Fl_Button(pos_x+(but_w+2)*2,3, but_w,but_h);
	btn->image(image_close_but);
	btn->box(box_type);
	btn->labeltype(FL_NORMAL_LABEL);
	//btn->align(FL_ALIGN_TOP | FL_ALIGN_INSIDE);
	btn->callback(on_close_btn);
	btn->when(FL_WHEN_RELEASE_ALWAYS);
	btn->user_data((void*)win);
	header->end();
	return header;
}

void resize_header(Fl_Window* header, int W){
	Fl_X::i(header)->origin_size.w = 0;//clear maximize flag !
	header->size(W, header->h());
}

void on_show_btn(Fl_Widget* pThis, void *param){
	if (param){
		Fl_Window* win = (Fl_Window*)param; 
		win->show();
	}
}

#include <FL/fl_draw.H>

#define MAX_ICON_WIDTH 100

Fl_Window* create_icon_win(int& W, int H, Fl_Window* win)
{
	Fl_Window* icon = new Fl_Window(0,0,W,H, ICON_LABEL); 
	icon->border(0);
	icon->box(FL_THIN_UP_BOX);
	icon->user_data((void*)win);
    const char* label = win->label();
    if (!label || strlen(label)>0)
      label = " ";
	Fl_Button *btn = new Fl_Button(2,1,W-4,H-2,label);
	btn->box(FL_NO_BOX);
	btn->down_box(FL_NO_BOX);
	icon->color(FL_DARK_BLUE, FL_DARK_BLUE);
	btn->labelcolor(FL_WHITE);
	btn->labelfont(FL_COURIER_BOLD);
	btn->labelsize(12);
	btn->user_data((void*)win);
	btn->callback(on_show_btn);
	//btn->when(FL_WHEN_RELEASE_ALWAYS);
	icon->resizable(btn);
	fl_font(FL_COURIER_BOLD, 12);
	int h=0;
	W=1000;
	char text[80]={0};
	strcpy(text,"@> ");
    if (strlen(btn->label()))
	  strncat(text,btn->label(),70);
	while(W >= MAX_ICON_WIDTH){
		fl_measure(text, W, h, -1);
		if (W >= MAX_ICON_WIDTH)
			text[strlen(text)-1]=0;
	}
	btn->copy_label(text);
	if (W < 30) W = 30;
//	if (W > MAX_ICON_WIDTH) W = MAX_ICON_WIDTH;
	W+=20;
	btn->align(FL_ALIGN_CENTER | FL_ALIGN_INSIDE);
	icon->size(W, icon->h());
	icon->end();
	return icon;
} 

void Fl_X::set_gc(int W, int H)
{
  int width,height;
  get_size(_xid, &width, &height);

  if (!win_gc){
    gc      = 0;
    win_gc  = create_gc(_xid);;
    clear_gc(win_gc,0,0,0,0);
    if (!w->border()){
      gc = win_gc;
      goto _l1;
    }
  }else{
    resize(_xid, width, height);
    clear_gc(win_gc,0,0,0,0);
  }
  if (w->border() && !w->parent()){
    DFBRectangle rect= {0,0,width,height};
    rect.x += WBORDER_W;
    rect.y += WHEADER_H;
    rect.w -= WBORDER_W*2;
    rect.h -= WHEADER_H+WBORDER_H;
    //redraw_header = true;
    if (gc)
      release_gc(gc);
    gc = create_gc(win_gc, &rect);
    if (header){
      resize_header(header, width);
      //Fl_X::i(header)->surface = win_surface;
    }else{
      header = create_header(width, WHEADER_H, w);
      header->parent(w);
      header->show();
    }
  }else{
    if (gc != win_gc)
      release_gc(gc);
    gc = win_gc;
  }
_l1:   
#if 0
	if (surface){
		release_surface(surface);
		if (surface == win_surface) 
			surface = win_surface = 0;
	}		
	if (win_surface){
		release_surface(win_surface);
		win_surface = 0;
	}
	if (w->border() && !w->parent()){
		w->end();//be sure that header is not a child of group 'w'
		win_surface = get_surface(xid); 
		clear_surface(win_surface, 0,0,0,0);
		int width,height;
		get_size(xid, &width, &height);
		if (W==-1){
			W = width-WBORDER_W*2;
			H = height-WHEADER_H+WBORDER_H;
		}
		DFBRectangle rect = { WBORDER_W,WHEADER_H,W,H };
		surface =get_sub_surface(win_surface, &rect);
  		if (header){
  			resize_header(header, width);
  		//	Fl_X::i(header)->surface = win_surface;
  		}else{
  			header = create_header(width, WHEADER_H, w);
			header->parent(w);
			header->show();
		}
	} else {
		if (header){
			header->hide();
			delete header;
			header = 0;
		}
		surface = win_surface = get_surface(xid);
  	}
#endif
  	for (Fl_X *wi = next; wi;){
   		Fl_Window* W = wi->w;
   		if (W->parent()) {
   			W->position(W->x(), W->y());
   		}
   		wi = wi->next;
	}
}

#define FL_CENTER_WINDOWS 1

Fl_X* Fl_X::make_xid(Fl_Window* w)
{ 
  Fl_Group::current(0); // get rid of very common user bug: forgot end()
  int X = w->x();
  int Y = w->y();
  int W = w->w();
  if (W <= 0) W = 1; // X don't like zero...
  int H = w->h();
  if (H <= 0) H = 1; // X don't like zero...
  if (!w->parent() && !Fl::grab()){
    // center windows in case window manager does not do anything:
#ifdef FL_CENTER_WINDOWS
    if (!(w->flags() & Fl_Window::FL_FORCE_POSITION)){
      w->x(X = Fl::x()+(Fl::w()-W)/2);
      w->y(Y = Fl::y()+(Fl::h()-H)/2);
    }
#endif// FL_CENTER_WINDOWS
    // force the window to be on-screen. Usually the X window manager
    // does this, but a few don't, so we do it here for consistency:
    int scr_x, scr_y, scr_w, scr_h;
    Fl::screen_xywh(scr_x, scr_y, scr_w, scr_h, X, Y);

    if (w->border()&& !w->parent()){
      H+=WHEADER_H+WBORDER_H;
      W+=WBORDER_W*2;
      X-=WBORDER_W;
      Y-=WHEADER_H;
    }
    // now insure contents are on-screen (more important than border):
    if (X+W > scr_x+scr_w) X = scr_x+scr_w-W;
    if (X < scr_x) X = scr_x;
    if (Y+H > scr_y+scr_h) Y = scr_y+scr_h-H;
    if (Y < scr_y) Y = scr_y;
  }else{
    if (w->parent()){
      //I dont create a new window because DirectFB does not support child windows
      Fl_X* parent = get_parent_xid(w);
      if (!parent)
        return 0;
      Fl_X* x = new Fl_X;
      //using description of the parent window
      memcpy(x, parent, sizeof(Fl_X));
      x->subWindow = 1;
      DFBRectangle rect = { X, Y , W, H };
      //create subsurface
      if (parent->header == w){
        x->gc = create_gc(parent->win_gc, &rect);
      }else{
        x->gc = create_gc(parent->gc, &rect);
      }
      //change window id to avoid finding out this child by parent id
      x->header = 0;
      x->xid = parent->xid | CHILD_ID;
      x->setwindow(w);
      //add to chain
      x->next = parent->next;
      parent->next = x;
      x->region = 0;
      x->first_draw = 1;
      w->redraw();
      return x;
    }
  }
#if 0  
  IDirectFBWindow *window;
  DFBDisplayLayerConfig layer_config;

  fl_display_layer->GetConfiguration(fl_display_layer, &layer_config);

  DFBWindowDescription desc;

#if 0  
  desc.flags = (DFBWindowDescriptionFlags)(DWDESC_CAPS | DWDESC_POSX |\
    DWDESC_POSY | DWDESC_WIDTH | DWDESC_HEIGHT | DWDESC_SURFACE_CAPS);

  desc.caps = DWCAPS_DOUBLEBUFFER;

  desc.surface_caps = DFBSurfaceCapabilities(DSCAPS_DOUBLE | MEMORY_TYPE);
#else
  desc.flags = (DFBWindowDescriptionFlags)(DWDESC_POSX |\
    DWDESC_POSY | DWDESC_WIDTH | DWDESC_HEIGHT | DWDESC_CAPS | DWDESC_SURFACE_CAPS);
  desc.caps = DWCAPS_ALPHACHANNEL;
  //desc.caps = DWCAPS_NONE;  
  desc.surface_caps = (DFBSurfaceCapabilities)(DSCAPS_PREMULTIPLIED);
#endif

  desc.posx = X;
  desc.posy = Y;
  desc.width = W;
  desc.height = H;

  window = create_window(&desc);
  set_opacity(window, 0);//hide the window
  DFBWindowOptions win_opt = (DFBWindowOptions)(DWOP_KEEP_SIZE |DWOP_ALPHACHANNEL);
  window->SetOptions(window, win_opt);
#endif  
  FL_DFB_Window* window = new FL_DFB_Window(X, Y, W, H);
  Fl_X* x = Fl_X::set_xid(w , window);
  x->subWindow = 0;

  x->set_gc(/*W,H*/);
#if 0  
  if (!fl_dfb_event_buffer){
    fl_dfb->CreateEventBuffer(fl_dfb, &fl_dfb_event_buffer);
    fl_display_layer->WarpCursor(fl_display_layer, X+W/2,Y+H/2);
    fl_dfb_event_buffer->CreateFileDescriptor(fl_dfb_event_buffer, &event_fd);
    Fl::add_fd(event_fd, POLLIN, fd_callback);
  }
  //window->AttachEventBuffer(window,fl_dfb_event_buffer);
#endif

  x->sendxjunk();

  x->cursors[0] = XCreateFontCursor(fl_display, (FL_CURSOR_ARROW-1)*2);
  XColor bg,fg;
  bg.red = bg.green = bg.blue = 255 << 8;
  fg.red = fg.green = fg.blue = 0;
  XRecolorCursor(fl_display, x->cursors[0], &fg, &bg);
  XDefineCursor(fl_display, x->xid, 0);

  int showit = 1;

  if (showit){
    w->set_visible();
    int old_event = Fl::e_number;
    w->handle(Fl::e_number = FL_SHOW); // get child windows to appear
    Fl::e_number = old_event;
    w->redraw();
  }
  return x;
}

extern Fl_Window *fl_xmousewin;
extern Fl_Window *fl_xfocus;
#include <FL/fl_draw.H>

#define sb_w 16
#define sb_h 16

void Fl_X::set_maximize(){
	if (maximize())
		return;
	origin_size.x = w->x();
	origin_size.y = w->y();
	int _w = w->w(),_h=w->h();
	if (w->border()){
		w->resize(Fl::x()+WBORDER_W, Fl::y()+WHEADER_H,Fl::w()-WBORDER_W*2, Fl::h()-WBORDER_H-WHEADER_H); 
	}else{
		w->resize(Fl::x(), Fl::y(),Fl::w(), Fl::h()); 
	}
	origin_size.w = _w;
	origin_size.h = _h;
	_TRACE("w %p %s origin xywh(%d %d %d %d)\n",w,w->label(),origin_size.x,origin_size.y,origin_size.w,origin_size.h);
       	
	//DFBWindowOptions win_opt = (DFBWindowOptions)(DWOP_KEEP_POSITION|DWOP_KEEP_SIZE|DWOP_ALPHACHANNEL);
	//_xid->SetOptions(_xid, win_opt);
}

void Fl_X::set_minimize(){
	if (!minimize()){
		int W = 80;
		int H = 20;
		Fl_Window* icon = create_icon_win(W,H,w);
		icon->position(current_icon_pos, Fl::h()-H);
		icon->user_data((void*)_xid);
		current_icon_pos += W;
		icon->show();
		minimize_=1;
		set_opacity(_xid, 0);
	}
}

void Fl_X::delete_icon()
{
  if (minimize()){
    int empty_pos = -1,shift = 0;
    for (Fl_X *wi = Fl_X::first; wi;){
      Fl_Window* W = wi->w;
      if (W->label()&&strcmp(W->label(),ICON_LABEL)==0){
        if (W->user_data()==_xid){
          empty_pos = W->x();
          shift = W->w();
          current_icon_pos -= W->w();
          W->hide();
          wi = Fl_X::first;
          continue;
        }else{
          if (empty_pos != -1){
            if (W->x() >= empty_pos){
              W->position(W->x()-shift, W->y());
              W->show();
            }
          }
        }
      }
      wi = wi->next;
    }
  }
  minimize_=0;
}

void Fl_X::restore_size(int X,int Y,int W,int H){
	if (maximize()){
		if (W == -1) W=origin_size.w;
		//else if (header) W+=WBORDER_W*2;
		if (H == -1) H=origin_size.h;
		//else if (header) H+=WBORDER_H+WHEADER_H;
		if (X == -1) X=origin_size.x;
		//else if (header) X-=WBORDER_W;
		if (Y == -1) Y=origin_size.y;
		//else if (header) Y-=WHEADER_H;
		origin_size.h = origin_size.w = 0;
		w->resize(X, Y, W, H);
		//DFBWindowOptions win_opt = (DFBWindowOptions)(DWOP_KEEP_SIZE|DWOP_ALPHACHANNEL);
		//_xid->SetOptions(_xid, win_opt);
	}
}

//function updates the window when I receive ENTER or LEAVE message
void Fl_X::enter_window(bool in){
  if (Fl::modal() && Fl::modal()!= w){
    if (!Fl::modal()->border() || in)
      return;
  }
  Fl_Widget *wdg = header;
  if (!wdg){
    if (!w->label()){
      if (w->parent()){
        wdg = get_parent_xid(w)->header;
      }else
        return;
    }else{
      if ((strcmp(w->label(), ICON_LABEL) == 0)){
        wdg = w->child(0);
      }else
        return;
    }
  }
  if (in)
    wdg->color(FL_DARK_BLUE);
  else
    wdg->color(12);
  wdg->redraw();
}

void Fl_X::draw_window_frame()
{
  if (!w->border() || w->parent())
    return;
  fl_gc     = win_gc;
  fl_window = (Window)win_gc;
  int W, H;
  get_size(win_gc, &W, &H);
  fl_clip_region(0);
  win_gc->surface->DrawRectangle(win_gc->surface,1,1,W-2,H-2);
  fl_draw_box(FL_UP_FRAME,0,0,W,H,FL_WHITE);
}

void Fl_X::flush()
{
  w->flush();
  if (w->damage() & FL_DAMAGE_ALL) { //FIXME I need some flag to restore subwindows ?
    Fl_X *p = next;
    while (p && (p->w->window()==w)){
      if (p->w->visible_r())
      {
        p->w->damage(FL_DAMAGE_ALL);
        p->flush();
      }
      p = p->next;
    }
  }
  w->make_current();
  if (first_draw){
    first_draw=0;
    if (w->border()&&header){
      draw_window_frame();
      header->flush();
      Fl_X::i(header)->first_draw = 0;
    }
    if ((w->label()==0)||(strcmp(w->label(),HEADER_LABEL))){//exclude header window
      if (!fl_show_iconic) //is not iconize ?
        set_opacity(_xid, w->opacity_);//display window
      else{
        set_opacity(_xid, 0);
        fl_show_iconic = 0;
        set_minimize();
      }
    }
    fl_flip(win_gc, 0);//flip the entire window surface
    _TRACE("flip the entire window\n");
    fl_show_cursor(_show_cursor);
    remove_updates(win_gc);
  }//else
  // FlipRegions(surface);
}

void Fl_X::release()
{
  if (fl_dfb == 0) return;
  if (Fl_X::first == 0){
    fl_deinit_dfb();
  }
}

#include "directfb_strings.h"

DirectFBSurfaceBlittingFlagsNames(dsblit_list);
DirectFBAccelerationMaskNames(dfxl_list);
DirectFBSurfaceDrawingFlagsNames(dsdraw_list);
DirectFBInputDeviceTypeFlagsNames(inpdev_list);

#  undef _TRACE
#  define _TRACE(p...) printf(p)
 
int screen_params(){
	get_size( fl_screen,&screen_width, &screen_height);
	fl_visual->byte_order = WORDS_BIGENDIAN; // MACROS
	static Visual cur_visual;
	fl_visual->visual = &cur_visual;
	pixf_ = get_pixel_format(fl_screen);
	lut_ = DFB_PIXELFORMAT_IS_INDEXED(pixf_);
	fl_visual->bytes_per_pixel = (pixf_ >> 20) & 7;
	fl_visual->bits_per_pixel = DFB_BITS_PER_PIXEL(pixf_);    
	switch (pixf_){
		case DSPF_LUT8:
			fl_visual->red_mask = fl_visual->blue_mask= fl_visual->green_mask=0;
            fl_visual->visual->red_mask = fl_visual->visual->blue_mask= fl_visual->visual->green_mask=0;
			fl_visual->colormap_size        = 256;
			fl_visual->depth                = 1;
			fl_visual->c_class              = PseudoColor;
			fl_visual->visual->map_entries  = 256;
			fl_visual->visual->bits_per_rgb = 8;//FIXME
			break;
		case DSPF_ARGB:	
		case DSPF_RGB32:	
		case DSPF_RGB24:
			fl_visual->red_mask   = fl_visual->visual->red_mask   = 0xFF0000;
			fl_visual->green_mask = fl_visual->visual->green_mask = 0x00FF00;
			fl_visual->blue_mask  = fl_visual->visual->blue_mask  = 0x0000FF;
			fl_visual->colormap_size        = 0;
			fl_visual->depth                = 1;
			fl_visual->c_class              = DirectColor;
			fl_visual->visual->map_entries  = 0;//FIXME
			fl_visual->visual->bits_per_rgb = 32;//FIXME 
 			break;
		case DSPF_RGB16:
			fl_visual->red_mask   = fl_visual->visual->red_mask   = 0xF800;
			fl_visual->green_mask = fl_visual->visual->green_mask = 0x07E0;
			fl_visual->blue_mask  = fl_visual->visual->blue_mask  = 0x001F;
			fl_visual->colormap_size        = 0;
			fl_visual->depth                = 1;
			fl_visual->c_class              = DirectColor;
			fl_visual->visual->map_entries  = 0;//FIXME
			fl_visual->visual->bits_per_rgb = 16;//FIXME	
			break;
        case DSPF_RGB332:
            fl_visual->red_mask   = fl_visual->visual->red_mask   = 0xE0;
            fl_visual->green_mask = fl_visual->visual->green_mask = 0x1C;
            fl_visual->blue_mask  = fl_visual->visual->blue_mask  = 0x03;
            fl_visual->colormap_size        = 0;
            fl_visual->depth                = 1;
            fl_visual->c_class              = DirectColor;
            fl_visual->visual->map_entries  = 0;//FIXME
            fl_visual->visual->bits_per_rgb = 8;//FIXME 
            break;
        case DSPF_ARGB4444:
            fl_visual->red_mask   = fl_visual->visual->red_mask   = 0x0F00;
            fl_visual->green_mask = fl_visual->visual->green_mask = 0x00F0;
            fl_visual->blue_mask  = fl_visual->visual->blue_mask  = 0x000F;
            fl_visual->colormap_size        = 0;
            fl_visual->depth                = 1;
            fl_visual->c_class              = DirectColor;
            fl_visual->visual->map_entries  = 0;//FIXME
            fl_visual->visual->bits_per_rgb = 16;//FIXME 
            break;            
		default: 
			Fl::fatal("Unsupported pixel format = %x!\n", pixf_);
	}
	TRACE("format '%s'\nbytes per pixel\t%d\nbits per pixel\t%d\nhas lut\t%d\n",\
            pixelformat_name((DFBSurfacePixelFormat)pixf_), fl_visual->bytes_per_pixel,fl_visual->bits_per_pixel,lut_);
	DFBGraphicsDeviceDescription dev_desc;
	fl_dfb->GetDeviceDescription(fl_dfb, &dev_desc);
	TRACE("Name\t'%s' , Vendor '%s'\n", dev_desc.name,  dev_desc.vendor);
	TRACE("Name\t'%s' , Vendor '%s'\n", dev_desc.driver.name,  dev_desc.driver.vendor);
	TRACE("Acceleration:\t");
	if (dev_desc.acceleration_mask == 0)
		TRACE("None");
	else{ 
		for (int i = 0; dfxl_list[i].mask; i++){
		  if (dev_desc.acceleration_mask & dfxl_list[i].mask)
		  	TRACE(" %s ",dfxl_list[i].name); 	
		}
	}
	TRACE("\nBlitting flags:\t");
	if (dev_desc.blitting_flags == 0)
		TRACE("None");
	else{ 
		for (int i = 0; dsblit_list[i].flag; i++){
		  if (dev_desc.blitting_flags & dsblit_list[i].flag)
		  	TRACE(" %s ",dsblit_list[i].name); 	
		}
	}
    TRACE("\nDrawing flags:\t");
	if (dev_desc.drawing_flags == 0)
		TRACE("None");
	else{ 
		for (int i = 0; dsdraw_list[i].flag; i++){
		  if (dev_desc.drawing_flags & dsdraw_list[i].flag)
		  	TRACE(" %s ",dsdraw_list[i].name); 	
		}
	}
	TRACE("\n");
	return dev_desc.acceleration_mask;
}

int deinit = 0; //flag for signal handler

void fl_deinit_dfb(void)
{
  if (!fl_dfb || deinit)
    return;

  deinit = 1;
  __TRACE("deinitializating DirectFB..\n");

  if (fl_dfb_event_buffer){
    fl_dfb_event_buffer->Release(fl_dfb_event_buffer);
    fl_dfb_event_buffer=0;
  }
  fl_free_dfbfonts();
  fl_free_cursors();

  __TRACE("print_gc_list\n");
  print_gc_list();

  __TRACE("free_gc_list\n");
  free_gc_list();

  if (fl_dfb_wm) {
    __TRACE("delete fl_dfb_wm\n");
  	delete fl_dfb_wm;
	fl_dfb_wm = 0;
  }
  
  // FIXME when I'll uncomment next line I'll have sigsegv on dfb->Release()... sometimes
  __TRACE("release fl_display_layer\n");
  if (fl_display_layer) {
      fl_display_layer->SetCooperativeLevel(fl_display_layer, DLSCL_SHARED);
      fl_display_layer->Release(fl_display_layer);
      fl_display_layer = 0;
  }

  __TRACE("release fl_dfb\n");
  if (fl_dfb) {
    fl_dfb->Release(fl_dfb);
    fl_dfb = 0;
  }
  __TRACE("DirectFB released\n");
}

int fl_init_dfb(int argc, char **argvv){
	static int beenhere = 0;
	if (!beenhere){
		beenhere = 1;
        DFBCHECK (DirectFBInit (&argc, (char***)&argvv));
		DFBCHECK (DirectFBCreate (&fl_dfb)); // Create the super interface.
	}
	return 0;
}

static void tick_cb(void*) {	
  	Fl::repeat_timeout(1.0, tick_cb);
}

void sigfunc(int sig){ 
     fprintf(stderr, "Signal '%d' received\n", sig); 
     if ((sig==SIGSEGV)||(sig==SIGINT)||(sig=SIGTERM)){
     	if (!deinit)
            fl_deinit_dfb();
     	exit (0); 
     }
}

DFBEnumerationResult myVideoModeCallback(int width,int height,int bpp,void *callbackdata){
	__TRACE("Video mode : width %d\theight %d\tbpp %d\n", width, height, bpp); 
	return DFENUM_OK;
}

DFBEnumerationResult myInputDeviceCallback(DFBInputDeviceID	device_id,DFBInputDeviceDescription desc,void *callbackdata){
	__TRACE("Input device ID =%d\t type= ",device_id);
	if (desc.type == 0){
		__TRACE("None");
	}else{
		for (int i = 0; inpdev_list[i].type; i++){
		  if (desc.type & inpdev_list[i].type)
		  	_TRACE(inpdev_list[i].name); 	
		}
	}
	__TRACE(	"\tname %s", desc.name); 
	__TRACE("\tvendor %s\n", desc.vendor); 
	if (desc.type & DIDTF_KEYBOARD){
		keyboard_id = device_id;
	}
	if (1 || !strcasecmp(desc.name,  "HID 3353:3713")) {
		__TRACE("create Intervideo Remote Control Event Filter\n");		
		if (!fl_dfb_eventFilter)
		    fl_dfb_eventFilter = Fl_DirectFB_EventFilter::createEventFilter(Fl_DirectFB_EventFilter::InterVideo);
	}
	return DFENUM_OK;
}

FL_EXPORT bool fl_application_exited() {
    return fl_exited;
}

FL_EXPORT void fl_exit_application(int code) {
    fl_exit_code = code;
    fl_exited = true;
}

FL_EXPORT void fl_close_display() {
    fl_deinit_dfb();
}

FL_EXPORT void fl_open_display(){
	if (!fl_dfb){
		TRACE("Starting DirectFB...\n");
#if 0//def VA_PLATFORM_EMMA3TL
        static char* argv[] = {"", "--dfb:no-hardware"};
        printf("**********FIXME: turn off directfb hardware accel on EMMA3TL\n");
        fl_init_dfb(2, argv);
#else    
		fl_init_dfb(0, 0);
#endif        
	}
	if (!fl_screen && fl_dfb){
		const char * enable_emma3pf_rc = getenv("ENABLE_EMMA3PF_RC_SUPPORT");
		if(enable_emma3pf_rc && strcasecmp("true", enable_emma3pf_rc) == 0){
			__TRACE("enable EMMA3PF RC support.\n");
			if (!fl_dfb_eventFilter)
				fl_dfb_eventFilter = Fl_DirectFB_EventFilter::createEventFilter(Fl_DirectFB_EventFilter::InterVideo);
		}
				
		fl_dfb->EnumVideoModes(fl_dfb,myVideoModeCallback,0);
		
		fl_dfb->EnumInputDevices(fl_dfb,myInputDeviceCallback,0);
        
   	  	DFBDisplayLayerConfig layer_config;
 
#ifdef VA_PLATFORM_EMMA3PF
    	fl_dfb->GetDisplayLayer(fl_dfb, DLID_PRIMARY, &fl_display_layer);
#else        
    	fl_dfb->GetDisplayLayer(fl_dfb, DLID_PRIMARY, &fl_display_layer);
#endif
        
        fl_display_layer->SetCooperativeLevel(fl_display_layer, DLSCL_ADMINISTRATIVE);
        //fl_display_layer->SetCooperativeLevel(fl_display_layer, DLSCL_EXCLUSIVE);
    	
	   	fl_display_layer->GetConfiguration(fl_display_layer, &layer_config);
     	//layer_config.buffermode = DLBM_BACKVIDEO;//DLBM_FRONTONLY;/*DLBM_BACKVIDEO;*///DLBM_WINDOWS;//
        layer_config.flags = (DFBDisplayLayerConfigFlags)(DLCONF_PIXELFORMAT | DLCONF_WIDTH | DLCONF_HEIGHT | DLCONF_BUFFERMODE);
     	layer_config.buffermode = DLBM_FRONTONLY;
     	//layer_config.buffermode = DLBM_BACKVIDEO;
        layer_config.width = 960;
        layer_config.height = 540;
        layer_config.pixelformat = DSPF_ARGB;
     	DFBCHECK(fl_display_layer->SetConfiguration(fl_display_layer, &layer_config));
        if (getenv("VA_LAYER_POSITION")) {
            double x = getenv("VA_LAYER_POSITION_X") ? atof(getenv("VA_LAYER_POSITION_X")) : 0.;
            double y = getenv("VA_LAYER_POSITION_Y") ? atof(getenv("VA_LAYER_POSITION_Y")) : 0.;
            double w = getenv("VA_LAYER_POSITION_W") ? atof(getenv("VA_LAYER_POSITION_W")) : 1.;
            double h = getenv("VA_LAYER_POSITION_H") ? atof(getenv("VA_LAYER_POSITION_H")) : 1.;
            DFBCHECK(fl_display_layer->SetScreenLocation(fl_display_layer, x, y, w, h));
        }
     	DFBCHECK(fl_display_layer->GetConfiguration(fl_display_layer, &layer_config));
        printf("layer information : \n");
        printf("\t flags = %x\n", layer_config.flags);
        printf("\t width = %d\n", layer_config.width);
        printf("\t height = %d\n", layer_config.height);
        printf("\t pixelformat = %x\n", layer_config.pixelformat);
        printf("\t buffermode = %d\n", layer_config.buffermode);
        printf("\t options = %x\n", layer_config.options);
        printf("\t source = %d\n", layer_config.source);
        printf("\t surface_caps = %x\n", layer_config.surface_caps);
       	   	
  		DFBGraphicsDeviceDescription  gdesc;
		fl_dfb->GetDeviceDescription(fl_dfb, &gdesc);
  		
        if (!fl_dfb_event_buffer){
            fl_dfb->CreateEventBuffer(fl_dfb, &fl_dfb_event_buffer);
            fl_dfb_event_buffer->CreateFileDescriptor(fl_dfb_event_buffer, &event_fd);
            Fl::add_fd(event_fd, POLLIN, fd_callback);
        }

        int showCursor = 1;
        if (getenv("FLTK_HIDE_CURSOR") && strcasecmp(getenv("FLTK_HIDE_CURSOR"), "true") == 0)
            showCursor = 0;
    	fl_display_layer->EnableCursor(fl_display_layer, showCursor);
    	fl_show_cursor(showCursor);
 
#if 0        
		DFBSurfaceDescription dsc; // A surface description is needed to create a surface.
		dsc.flags = DFBSurfaceDescriptionFlags(DSDESC_CAPS /*| DSDESC_PIXELFORMAT*/);
		dsc.caps=(DFBSurfaceCapabilities)(DSCAPS_PRIMARY|DSCAPS_FLIPPING/*|DSCAPS_STATIC_ALLOC */);
		if (gdesc.acceleration_mask){
			dsc.caps = (DFBSurfaceCapabilities)(dsc.caps | DSCAPS_VIDEOONLY);
		}
		fl_screen = create_gc(&dsc);
#else
        fl_screen = create_gc(fl_display_layer);
#endif
		screen_params();
        //clear_gc(fl_screen, 0, 0, 0, 255);
        //flip_gc( fl_screen, 0,  DSFLIP_NONE);
        //clear_gc(fl_screen, 0, 0, 0, 255);

#	if !USE_COLORMAP		
		int pf = pixel_format();
		if (pf != DSPF_ARGB && pf != DSPF_RGB32 && pf != DSPF_RGB16)
			Fl::fatal("FLTK need 'RGB32' or 'ARGB' modes!"); 
#	endif		
		
#  	if USE_COLORMAP     	
		fl_colormap = XCreateColormap(fl_display);
		init_xcolormap();
		XFlushColormap(fl_display, fl_colormap);
		BG_COLOR_IDX = fl_xpixel(Fl_Color(BG_COLOR_RGB32 << 8));
		MASK_COLOR_IDX = fl_xpixel(MASK_COLOR_RED , MASK_COLOR_GREEN, MASK_COLOR_BLUE);
#	endif		
     	Fl::add_timeout(1.0, tick_cb);

        fl_dfb_wm = new FL_DFB_WindowManager();

     	//signal(SIGSEGV, sigfunc);
     	//signal(SIGINT, sigfunc);
     	//signal(SIGTERM, sigfunc);
     	atexit(fl_deinit_dfb); 
   	}
}

/////////////////
#if DIRECTFB_MAJOR_VERSION == 0
void
dfb_updates_init( DFBUpdates *updates,
                  DFBRegion  *regions,
                  int         max_regions )
{
     D_ASSERT( updates != NULL );
     D_ASSERT( regions != NULL );
     D_ASSERT( max_regions > 0 );

     updates->regions     = regions;
     updates->max_regions = max_regions;
     updates->num_regions = 0;

     D_MAGIC_SET( updates, DFBUpdates );
}

void
dfb_updates_add( DFBUpdates      *updates,
                 const DFBRegion *region )
{
     int i;

     D_MAGIC_ASSERT( updates, DFBUpdates );
     DFB_REGION_ASSERT( region );
     D_ASSERT( updates->regions != NULL );
     D_ASSERT( updates->num_regions >= 0 );
     D_ASSERT( updates->num_regions <= updates->max_regions );

     if (updates->num_regions == 0) {
          updates->regions[0]  = updates->bounding = *region;
          updates->num_regions = 1;

          return;
     }

     for (i=0; i<updates->num_regions; i++) {
          DFBRegion inter = updates->regions[i];

          if (dfb_region_region_intersect( &inter, region )) {
               DFBRegion uni = updates->regions[i];

               dfb_region_region_union( &uni, region );

               updates->regions[i] = uni;

               dfb_region_region_union( &updates->bounding, &uni );

               return;
          }
     }

     if (updates->num_regions == updates->max_regions) {
          dfb_region_region_union( &updates->bounding, region );

          updates->regions[0]  = updates->bounding;
          updates->num_regions = 1;
     }
     else {
          updates->regions[updates->num_regions++] = *region;

          dfb_region_region_union( &updates->bounding, region );
     }
}

void
dfb_updates_stat( DFBUpdates *updates,
                  int        *ret_total,
                  int        *ret_bounding )
{
     int i;

     D_MAGIC_ASSERT( updates, DFBUpdates );
     D_ASSERT( updates->regions != NULL );
     D_ASSERT( updates->num_regions >= 0 );
     D_ASSERT( updates->num_regions <= updates->max_regions );

     if (updates->num_regions == 0) {
          if (ret_total)
               *ret_total = 0;

          if (ret_bounding)
               *ret_bounding = 0;

          return;
     }

     if (ret_total) {
          int total = 0;

          for (i=0; i<updates->num_regions; i++) {
               const DFBRegion *r = &updates->regions[i];

               total += (r->x2 - r->x1 + 1) * (r->y2 - r->y1 + 1);
          }

          *ret_total = total;
     }

     if (ret_bounding)
          *ret_bounding = (updates->bounding.x2 - updates->bounding.x1 + 1) *
                          (updates->bounding.y2 - updates->bounding.y1 + 1);
}
#endif
/////////////////

