//
// "$Id: _arcpoly.cxx,v 1.2 2006-12-06 11:37:23 nikego Exp $"
//
// DirectFB-specific code for the Fast Light Tool Kit (FLTK).
//
// Copyright 1998-2005 by Bill Spitzak and others.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems on the following page:
//
//     http://www.fltk.org/str.php
//
// This file contains arc drawing code for fltk which is always linked
// in.  Search other files for "DIRECTFB" or filenames ending in _directfb.cxx
// for other system-specific code.
//
// This file was taken from  Microwindows (Nano-X Window System) projects
//  http://www.microwindows.org
//  Edited by Nikita Egorov (nikego@gmail.com)  
//
//
// Copyright (c) 2000-2001 Greg Haerr <greg@censoft.com>
// Portions Copyright (c) 2002 by Koninklijke Philips Electronics N.V.
//
// Device-independent arc, pie and ellipse routines.
// GdArc is integer only and requires start/end points.
// GdArcAngle requires floating point and uses angles.
// GdArcAngle uses qsin() and qcos() instead of sin() / cos() 
// so no math lib needed.
//
// Portions Copyright (c) 1991 David I. Bell
//
// Arc line clipping and integer qsin/qcos routines used by permission:
// Copyright (C) 1997-1998 by Eero Tamminen
// Bugfixed by Greg Haerr
 

/* NIKITA: FIXME DirectFB does not need clipping */
#define NOCLIPPING 1

/* NIKITA: FIXME FLTK has only solid fill */
#define ONLYSOLIDFILL 1

typedef GC PSD; //short name for surface's class

/* variables for Nano X */

int gr_fillmode;


void drawpoint(GC gc,int x,int y){
  	gc->surface->DrawLine(gc->surface, x,y,x,y);
}

void drawrow(GC gc,int x1,int x2,int y){
 	gc->surface->DrawLine(gc->surface, x1, y ,x2, y);
}

#define GdLine(psd, _x1, _y1, _x2, _y2, _b) fl_line(_x1, _y1, _x2, _y2) 
//stubs 
inline int GdFixCursor(GC ){return 0;} 
inline int GdSetDash(unsigned long*, int*){return 0;} 

#ifdef HAVEFLOAT

#include <math.h>
#define qcos	QCOS
#define qsin	QSIN
typedef double	FLOAT;

FLOAT QCOS(FLOAT a){
	return cos(a * M_PI / 180.);
}

FLOAT QSIN(FLOAT a){
	return sin(a * M_PI / 180.);
}
#endif /* HAVEFLOAT*/

/**
 * Draw an arc or pie, angles are specified in 64th's of a degree.
 * This function requires floating point, use GdArc for integer only.
 *
 * @param psd Destination surface.
 * @param x0 Center of arc (X co-ordinate).
 * @param y0 Center of arc (Y co-ordinate).
 * @param rx Radius of arc in X direction.
 * @param ry Radius of arc in Y direction.
 * @param angle1 Start of arc.  In 64ths of a degree, anticlockwise from
 * the +x axis.
 * @param angle2 End of arc.  In 64ths of a degree, anticlockwise from
 * the +x axis.
 * @param type Type of arc:
 * ARC is a curved line.
 * ARCOUTLINE is a curved line plus straight lines joining the ends
 * to the center of the arc.
 * PIE is a filled shape, like a section of a pie chart.
 */
#if 0 //nikita 
void
GdArcAngle(PSD psd, short x0, short y0, short rx, short ry,
	short angle1, short angle2, int type)
{
#if HAVEFLOAT
	short	ax, ay, bx, by;
	FLOAT	a, b;

	/* calculate pie edge offsets from center to the ellipse rim */
	ax = qcos(angle1/64.) * rx;
	bx = qcos(angle2/64.) * rx;

	a = -qsin(angle1/64.);
	b = -qsin(angle2/64.);
	ay = a * ry;
	by = b * ry;

	/* fix nxlib error with very small angles becoming whole circle*/
	if (ax == bx && ay == by && angle1 != angle2)
		return;

	/* call integer routine*/
	GdArc(psd, x0, y0, rx, ry, ax, ay, bx, by, type);
#endif /*HAVEFLOAT*/
}
#endif
/* argument holder for pie, arc and ellipse functions*/
typedef struct {
	PSD	psd;
	short	x0, y0;
	short	rx, ry;
	short	ax, ay;
	short	bx, by;
	int	adir, bdir;
	int	type;
} SLICE;

/*
 * Clip a line segment for arc or pie drawing.
 * Returns 0 if line is clipped or on acceptable side, 1 if it's vertically
 * on other side, otherwise 3.
 */
static int
clip_line(SLICE *slice, short xe, short ye, int dir, short y, short *x0,
	short *x1)
{
#if 0
	/*
	 * kluge: handle 180 degree case
	 */
	if (y >= 0 && ye == 0) {
		/* bottom 180*/
		if (slice->adir < 0) {
			if (slice->ay || slice->by)
				return 1;
			if (slice->ax == -slice->bx)
				return 0;
		}
		return 3;
	}
#endif
	/* hline on the same vertical side with the given edge? */
	if ((y >= 0 && ye >= 0) || (y < 0 && ye < 0)) {
		short x;

		if (ye == 0) x = xe; else
		x = (short)(long)xe * y / ye;

		if (x >= *x0 && x <= *x1) {
			if (dir > 0)
				*x0 = x;
			else
				*x1 = x;
			return 0;
		} else {
			if (dir > 0) {
				if (x <= *x0)
					return 0;
			} else {
				if (x >= *x1)
					return 0;
			}
		}
		return 3;
	}
	return 1;
}

/* relative offsets, direction from left to right. */
/* Mode indicates if we are in filil mode (1) or line mode (0) */

static void
draw_line(SLICE *slice, short x0, short y, short x1, int mode)
{
	int	dbl = (slice->adir > 0 && slice->bdir < 0);
	int 	discard, ret;
	short	x2 = x0, x3 = x1;

	if (y == 0) {
		if (slice->type != PIE){
			goto label_1; /*FIXME 
			if we doing the "return" command we have'nt two points on X axis */
			/*return;*/
		}		
		/* edges on different sides */
		if ((slice->ay <= 0 && slice->by >= 0) ||
		    (slice->ay >= 0 && slice->by <= 0)) {
			if (slice->adir < 0)  {
				if (x1 > 0)
					x1 = 0;
			}
			if (slice->bdir > 0) {
				if (x0 < 0)
					x0 = 0;
			}
		} else {
			if (!dbl) {
				/* FIXME leaving in draws dot in center*/
#ifndef ONLYSOLIDFILL
			        if (gr_fillmode != FILL_SOLID && mode) 
				  ts_drawpoint(slice->psd, slice->x0, slice->y0);
				else 
#endif				
				  drawpoint(slice->psd, slice->x0, slice->y0);
				return;
			}
		}
#ifndef ONLYSOLIDFILL		
		if (gr_fillmode != FILL_SOLID && mode)
		  ts_drawrow(slice->psd, slice->x0 + x0, slice->x0 + x1, slice->y0);
		else
#endif		
		  drawrow(slice->psd, slice->x0 + x0, slice->x0 + x1, slice->y0);
		return;
	}
label_1:
	/* clip left edge / line */
	ret = clip_line(slice, slice->ax, slice->ay, slice->adir, y, &x0, &x1);

	if (dbl) {
		if (!ret) {
			/* edges separate line to two parts */
#ifndef ONLYSOLIDFILL			
		        if (gr_fillmode != FILL_SOLID && mode)
			  ts_drawrow(slice->psd, slice->x0 + x0, slice->x0 + x1,
				     slice->y0 + y);
			else
#endif			
			  drawrow(slice->psd, slice->x0 + x0, slice->x0 + x1,
				  slice->y0 + y);

			x0 = x2;
			x1 = x3;
		}
	} else {
		if (ret > 1) {
			return;
		}
	}

	discard = ret;
	ret = clip_line(slice, slice->bx, slice->by, slice->bdir, y, &x0, &x1);

	discard += ret;
	if (discard > 2 && !(dbl && ret == 0 && discard == 3)) {
		return;
	}
	if (discard == 2) {
		/* line on other side than slice */
		if (slice->adir < 0 || slice->bdir > 0) {
			return;
		}
	}
#ifndef ONLYSOLIDFILL	
	if (gr_fillmode != FILL_SOLID && mode)
	  ts_drawrow(slice->psd, slice->x0 + x0, slice->x0 + x1, slice->y0 + y);
	else
#endif	
	  drawrow(slice->psd, slice->x0 + x0, slice->x0 + x1, slice->y0 + y);
}

/*
 * draw one line segment or set of points, called from drawarc routine
 *
 * Note that this is called for all rows in one quadrant of the ellipse.
 * It mirrors vertically & horizontally to get the entire ellipse.
 *
 * It passes on co-ordinates for the *entire* ellipse - for pie and
 * arc, clipping is done later to ensure that only the requested angle
 * gets drawn.
 */
static void
drawarcsegment(SLICE *slice, short xp, short yp, int drawon)
{
	unsigned long dm = 0;
	int dc = 0;

	switch (slice->type) {
	case ELLIPSEFILL:
		/* draw ellipse fill segment*/
                /* First, save the dash settings, because we don't want to use them here */
#ifndef ONLYSOLIDFILL
	  if (gr_fillmode != FILL_SOLID) {
	    ts_drawrow(slice->psd, slice->x0-xp, slice->x0+xp, slice->y0-yp);
	    ts_drawrow(slice->psd, slice->x0-xp, slice->x0+xp, slice->y0+yp);
	  }
	  else 
#endif	  
	  {
	    GdSetDash(&dm, &dc); /* Must turn off the dash settings because of drawrow() */
	    drawrow(slice->psd, slice->x0-xp, slice->x0+xp, slice->y0-yp);
	    drawrow(slice->psd, slice->x0-xp, slice->x0+xp, slice->y0+yp);
	    GdSetDash(&dm, &dc);
	  }

	  return;

	case ELLIPSE:
	  if (!drawon) return;
		/* set four points symmetrically situated around a point*/
		drawpoint(slice->psd, slice->x0 + xp, slice->y0 + yp);
		drawpoint(slice->psd, slice->x0 - xp, slice->y0 + yp);
		drawpoint(slice->psd, slice->x0 + xp, slice->y0 - yp);
		drawpoint(slice->psd, slice->x0 - xp, slice->y0 - yp);
		return;

	case PIE:
		/* draw top and bottom halfs of pie*/
	    if (gr_fillmode == FILL_SOLID) GdSetDash(&dm, &dc);
		draw_line(slice, -xp, -yp, +xp, 1);
		draw_line(slice, -xp, +yp, +xp, 1);
		if (gr_fillmode == FILL_SOLID) GdSetDash(&dm, &dc);
		return;

	default:	/* ARC, ARCOUTLINE*/
		/* set four points symmetrically around a point and clip*/
		draw_line(slice, +xp, +yp, +xp, 0);
		draw_line(slice, -xp, +yp, -xp, 0);
		draw_line(slice, +xp, -yp, +xp, 0);
		draw_line(slice, -xp, -yp, -xp, 0);
		return;
	}
}

/* General routine to plot points on an arc.  Used by arc, pie and ellipse*/
static void
drawarc(SLICE *slice)
{
	short xp, yp;		/* current point (based on center) */
	short rx, ry;
	long Asquared;		/* square of x semi axis */
	long TwoAsquared;
	long Bsquared;		/* square of y semi axis */
	long TwoBsquared;
	long d;
	long dx, dy;

	int bit  = 0;
	int drawon = 1;

	rx = slice->rx;
	ry = slice->ry;

	xp = 0;
	yp = ry;
	Asquared = rx * rx;
	TwoAsquared = 2 * Asquared;
	Bsquared = ry * ry;
	TwoBsquared = 2 * Bsquared;
	d = Bsquared - Asquared * ry + (Asquared >> 2);
	dx = 0;
	dy = TwoAsquared * ry;
#ifndef ONLYSOLIDFILL
	if (gr_fillmode != FILL_SOLID)
	  set_ts_origin(slice->x0 - rx, slice->y0 - ry);
#endif
	while (dx < dy) {

		/*
		 * Only draw if one of the following conditions holds:
		 * - We're drawing an outline - i.e. slice->type is
		 *   not PIE or ELLIPSEFILL
		 * - We're about to move on to the next Y co-ordinate
		 *   (i.e. we're drawing a filled shape and we're at
		 *   the widest point for this Y co-ordinate).
		 *   This is the case if d (the error term) is >0
		 * Otherwise, we draw multiple times, which messes up
		 * with SRC_OVER or XOR modes.
		 */
		if ((d > 0) || ((slice->type != PIE) && (slice->type != ELLIPSEFILL))) {
			if (slice->psd->dashcount) {
				drawon = (slice->psd->dashmask & (1 << bit)) ? 1 : 0;
				bit = (bit + 1) % slice->psd->dashcount;
			} else
				drawon = 1;

			drawarcsegment(slice, xp, yp, drawon);
		}

		if (d > 0) {
			yp--;
			dy -= TwoAsquared;
			d -= dy;
		}
		xp++;
		dx += TwoBsquared;
		d += (Bsquared + dx);
	}

	d += ((3L * (Asquared - Bsquared) / 2L - (dx + dy)) >> 1);

	while (yp >= 0) {
	        if (slice->psd->dashcount) {
	          drawon = (slice->psd->dashmask & (1 << bit)) ? 1 : 0;
		  bit = (bit + 1) % slice->psd->dashcount;
		}
	        else drawon = 1;

		drawarcsegment(slice, xp, yp, drawon);
		if (d < 0) {
			xp++;
			dx += TwoBsquared;
			d += dx;
		}
		yp--;
		dy -= TwoAsquared;
		d += (Asquared - dy);
	}

}

/**
 * Draw an arc or pie using start/end points.
 * Integer only routine.  To specify start/end angles,
 * use GdArcAngle, which requires floating point.
 *
 * @param psd Destination surface.
 * @param x0 Center of arc (X co-ordinate).
 * @param y0 Center of arc (Y co-ordinate).
 * @param rx Radius of arc in X direction.
 * @param ry Radius of arc in Y direction.
 * @param ax Start of arc (X co-ordinate).
 * @param ay Start of arc (Y co-ordinate).
 * @param bx End of arc (X co-ordinate).
 * @param by End of arc (Y co-ordinate).
 * @param type Type of arc:
 * ARC is a curved line.
 * ARCOUTLINE is a curved line plus straight lines joining the ends
 * to the center of the arc.
 * PIE is a filled shape, like a section of a pie chart.
 */
void
Arc(GC gc, short x0, short y0, short rx, short ry,
	short ax, short ay, short bx, short by, int type)
{
	//direct_draw psd(surf);
	//psd.setcolor(color_index_);
	PSD psd = gc;
	short	adir, bdir;
	SLICE	slice;
	//_TRACE("rx %d ry %d\n", rx,ry);
		
	if (rx < 0 || ry < 0)
		return;
		
	update_region(psd,x0-rx,y0-ry,x0+rx,y0+ry);	
	
	if (rx==0) {gc->surface->DrawLine(gc->surface,x0,y0-ry,x0,y0+ry);return;}
	if (ry==0) {gc->surface->DrawLine(gc->surface,x0-rx,y0,x0+rx,y0);return;}

	
	/*
	 * Calculate right/left side clipping, based on quadrant.
	 * dir is positive when right side is filled and negative when
	 * left side is to be filled.
	 *
	 * >= 0 is bottom half
	 */
	if (ay >= 0)
		adir = 1;
	else
		adir = -1;

	if (by >= 0)
		bdir = -1;
	else
		bdir = 1;

	/*
	 * The clip_line routine has problems around the 0 and
	 * 180 degree axes.
	 * This <fix> is required to make the clip_line algorithm
	 * work.  Getting these routines to work for all angles is
	 * a bitch.  And they're still buggy.  Doing this causes
	 * half circles to be outlined with a slightly bent line
	 * on the x axis. FIXME
	 */
	if (ay == 0) ++ay;
	if (by == 0) ++by;

	/* swap rightmost edge first */
	if (bx > ax) {
		short swap;

		swap = ax;
		ax = bx;
		bx = swap;

		swap = ay;
		ay = by;
		by = swap;

		swap = adir;
		adir = bdir;
		bdir = swap;
	}
#ifndef NOCLIPPING
	/* check for entire area clipped, draw with per-point clipping*/
	if (GdClipArea(psd, x0-rx, y0-ry, x0+rx, y0+ry) == CLIP_INVISIBLE)
		return;
#endif
	slice.psd = psd;
	slice.x0 = x0;
	slice.y0 = y0;
	slice.rx = rx;
	slice.ry = ry;
	slice.ax = ax;
	slice.ay = ay;
	slice.bx = bx;
	slice.by = by;
	slice.adir = adir;
	slice.bdir = bdir;
	slice.type = type;

	drawarc(&slice);
	
	if (type & OUTLINE) {
		/* draw two lines from rx,ry to arc endpoints*/
		GdLine(psd, x0, y0, x0+ax, y0+ay, TRUE);
		GdLine(psd, x0, y0, x0+bx, y0+by, TRUE);
	}

	GdFixCursor(psd);
	//psd.end_draw();
}

/**
 * Draw an ellipse using the current clipping region and foreground color.
 * This draws in the outline of the ellipse, or fills it.
 * Integer only routine.
 *
 * @param psd Destination surface.
 * @param x Center of ellipse (X co-ordinate).
 * @param y Center of ellipse (Y co-ordinate).
 * @param rx Radius of ellipse in X direction.
 * @param ry Radius of ellipse in Y direction.
 * @param fill Nonzero for a filled ellipse, zero for an outline.
 */
void
GdEllipse(GC gc, short x, short y, short rx, short ry, bool fill)
{
	//direct_draw psd(surf);
	//psd.setcolor(color_index_);
	PSD psd = gc;
	SLICE	slice;

	if (rx < 0 || ry < 0)
		return;

	/* Check if the ellipse bounding box is either totally visible
	 * or totally invisible.  Draw with per-point clipping.
	 */
#ifndef NOCLIPPING	 
	switch (GdClipArea(psd, x - rx, y - ry, x + rx, y + ry)) {
	case CLIP_VISIBLE:
		/*
		 * For size considerations, there's no low-level ellipse
		 * draw, so we've got to draw all ellipses
		 * with per-point clipping for the time being
		psd->DrawEllipse(psd, x, y, rx, ry, fill, gr_foreground);
		GdFixCursor(psd);
		return;
		 */
		break;

	case CLIP_INVISIBLE:
		return;
  	}
#endif
	slice.psd = psd;
	slice.x0 = x;
	slice.y0 = y;
	slice.rx = rx;
	slice.ry = ry;
	slice.type = fill? ELLIPSEFILL: ELLIPSE;
	/* other elements unused*/

	drawarc(&slice);

	GdFixCursor(psd);
//	psd.end_draw();
}
/************************************************************************/
#define max_columns 64
#define max_lines   2048

typedef struct column_item {
     DFBSpan* list;
     int first_y;
     int last_y;
};

static column_item spanslist[max_columns] = {{0}};
 
/*
 * flush spans list to screen
 */  
static int
flush_spans(GC gc)
{
  // flash all spans list and release memory
  int i = 0;
  column_item* item = spanslist;
  for (; i < max_columns; i++)
  {
    if (item->list)
    {
      int s = item->first_y, j;// 's' points to first span
      for (j = item->first_y; j <= item->last_y; j++)
      {
        if (item->list[j].w == 0)//it's empty chunk
        {
          if (j - s)// have we a number of spans to flush?
            gc->surface->FillSpans(gc->surface, s, item->list+s, j-s);
          s = j + 1;// shift our 's' to next non empty span
        }
      }
      if (j - s) // after ending of cycle test for remaining spans
        gc->surface->FillSpans(gc->surface, s, item->list+s, j-s);

      Xfree(item->list);
      item->list = 0;
      item++; //next list
      continue;
    }
    break;
  }
  return 0;
}

//add new horz line into lists of spans
static int
add_span(GC gc, int y, int x, int w)
{
  if (y < 0 || y > max_lines)
    return -1;
  column_item* item = spanslist;
  int i = 0;
  do {
    if (item->list == 0)
    {
      item->list = (DFBSpan*)Xcalloc(sizeof(DFBSpan) , max_lines);
      item->first_y = 10000; //init min and max values of 'y'
      item->last_y  =-10000;
    }
    if (item->list[y].w == 0)//empty span with our 'y'
    {
      item->list[y].x = x; //put 'x' and width
      item->list[y].w = w;
      if (item->first_y > y) //if it's necessary correct min and max values of 'y'
        item->first_y = y;
      if (item->last_y < y)
        item->last_y = y;
      return 0; //success exit
    }
    item++;//step to next list
  }while (++i < max_columns);
  //oh-oh! something is wrong
  return -1;
}

#if EDGEPOLYFILL    /* irregular polygon fill, uses edge table, malloc, qsort*/

/************************************************************/

/*
 * Fill a polygon in the foreground color, applying clipping if necessary.
 * The last point may be a duplicate of the first point, but this is
 * not required.
 * Note: this routine correctly draws convex, concave, regular, 
 * and irregular polygons.
 */
#define swap(a,b) do { a ^= b; b ^= a; a ^= b; } while (0)

typedef struct {
    int     x1, y1, x2, y2;
#if HAVEFLOAT
    double  x, m;
#else
    int     cx, fn, mn, d;
#endif
} edge_t;

static inline int 
edge_cmp(edge_t *lp, edge_t *rp)
{
     /* if the minimum y values are different, sort on minimum y */
    if (lp->y1 != rp->y1)
        return (lp->y1 > rp->y1);

    /* if the current x values are different, sort on current x */
    return (lp->x > rp->x);
}

static inline void 
swap_edges(edge_t* a, edge_t* b){ 
    edge_t t = *a;
    *a = *b;
    *b = t;
}

/*
 * code of these sort functions was taken from "Programming Pearls"
 * http://netlib.bell-labs.com/cm/cs/pearls/sort.cpp
 */
 
static inline void 
isort3(edge_t *x, int n)
{   int i, j;
    edge_t t;
    for (i = 1; i < n; i++) {
        t = x[i];
        for (j = i; j > 0 && edge_cmp(x+j-1,&t); j--)
            x[j] = x[j-1];
        x[j] = t;
    }
}
/* qsort3 + randomization + isort small subarrays + swap inline */
static int cutoff = 50;

static void qsort4(edge_t *x, int l, int u)
{   int i, j;
    edge_t t, temp;
    if (u - l < cutoff)
        return;
    //swap_edges(x+l, x+randint(l, u));//FIXME - it hangs up...
    t = x[l];
    i = l;
    j = u+1;
    for (;;) {
        do i++; while (i <= u && edge_cmp(&t,x+i));
        do j--; while (edge_cmp(x+j,&t));
        if (i > j)
            break;
        temp = x[i]; x[i] = x[j]; x[j] = temp;
    }
    swap_edges(x+l, x+j);
    qsort4(x,l, j-1);
    qsort4(x,j+1, u);
}

/* for big polygons (where egdes number > cutoff) we are using quick sort */
#define SORT1(ptr,n) qsort4(ptr,0,n-1);isort3(ptr,n)

/* otherwise,  insertion sort works faster */
#define SORT2(ptr,n) isort3(ptr,n)

/**
 * Draw a filled polygon.
 *
 * @param psd Drawing surface.
 * @param count Number of points in polygon.
 * @param pointtable The array of points.
 */
void
FillPoly(Drawable d, GC gc, int count, POINT_f* pointtable){

    Surface surface = ISGC(d);
    if (!surface)
      surface = gc;
  

    edge_t *get;        /* global edge table */
    int     nge = 0;    /* num global edges */
    int     cge = 0;    /* cur global edge */

    edge_t *aet;        /* active edge table */
    int     nae = 0;    /* num active edges */

    int     i, y;

    if (count < 3) {
        /* error, polygons require at least three edges (a triangle) */
        return;
    }
    get = (edge_t *) calloc(count, sizeof(edge_t));
    aet = (edge_t *) calloc(count, sizeof(edge_t));

    if ((get == 0) || (aet == 0)) {
        /* error, couldn't allocate one or both of the needed tables */
        if (get)
            free(get);
        if (aet)
            free(aet);
        return;
    }
    /* setup the global edge table */
    for (i = 0; i < count; ++i) {
        get[nge].x1 = (int)(pointtable[i].x+.5);
        get[nge].y1 = (int)(pointtable[i].y+.5);
        get[nge].x2 = (int)(pointtable[(i + 1) % count].x+.5);
        get[nge].y2 = (int)(pointtable[(i + 1) % count].y+.5);
        if (get[nge].y1 != get[nge].y2) {
            if (get[nge].y1 > get[nge].y2) {
                swap(get[nge].x1, get[nge].x2);
                swap(get[nge].y1, get[nge].y2);
            }
#if HAVEFLOAT
            get[nge].x = get[nge].x1;
            get[nge].m = get[nge].x2 - get[nge].x1;
            get[nge].m /= get[nge].y2 - get[nge].y1;
#else
            get[nge].cx = get[nge].x1;
            get[nge].mn = get[nge].x2 - get[nge].x1;
            get[nge].d = get[nge].y2 - get[nge].y1;
            get[nge].fn = get[nge].mn / 2;
#endif
            ++nge;
        }
    }

    SORT1(get,nge);
    
    /* start with the lowest y in the table */
    y = get[0].y1;

    do {
        /* add edges to the active table from the global table */
        while ((nge > 0) && (get[cge].y1 == y)) {
            aet[nae] = get[cge++];
            --nge;
            aet[nae++].y1 = 0;
        }
        SORT2(aet,nae);
        /* using odd parity, render alternating line segments */
        for (i = 1; i < nae; i += 2) {
#if HAVEFLOAT
            int     l = (int)(aet[i - 1].x + 0.5);//+0.5 - Nikita
            int     r = (int)(aet[i].x + 0.5);//+0.5 - Nikita
#else
            int     l = (int)aet[i - 1].cx;
            int     r = (int)aet[i].cx;
#endif
            if (r > l){
                add_span(surface, y, l, r-l);
            }   
        }

        /* prepare for the next scan line */
        ++y;

        /* remove inactive edges from the active edge table */
        /* or update the current x position of active edges */
        for (i = 0; i < nae; ++i) {
            if (aet[i].y2 == y)
                aet[i--] = aet[--nae];
            else {
#if HAVEFLOAT
                aet[i].x += aet[i].m;
#else
                aet[i].fn += aet[i].mn;
                if (aet[i].fn < 0) {
                    aet[i].cx += aet[i].fn / aet[i].d - 1;
                    aet[i].fn %= aet[i].d;
                    aet[i].fn += aet[i].d;
                }
                if (aet[i].fn >= aet[i].d) {
                    aet[i].cx += aet[i].fn / aet[i].d;
                    aet[i].fn %= aet[i].d;
                }
#endif
            }
        }
        /* keep doing this while there are any edges left */
    } while ((nae > 0) || (nge > 0));
    
    flush_spans(surface);
    /* all done, free the edge tables */
    free(get);
    free(aet);

    GdFixCursor(surface);
    //_TRACE("polygon reg(%d %d %d %d)\n",minx,miny,maxx,maxy);
    //update_region(surf,minx,miny,maxx,maxy);
}

#endif /* EDGEPOLYFILL*/

