//
// "$Id: _colormap.cxx,v 1.2 2006-10-31 09:31:14 nikego Exp $"
//
// DirectFB-specific code for the Fast Light Tool Kit (FLTK).
//
// Copyright 1998-2005 by Bill Spitzak and others.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems on the following page:
//
//     http://www.fltk.org/str.php
//
// This file contains colormap-specific code for fltk which is always linked
// in.  Search other files for "DIRECTFB" or filenames ending in _directfb.cxx
// for other system-specific code.
//
// This file was taken from  Microwindows (Nano-X Window System) projects
// http://www.microwindows.org
// Edited by Nikita Egorov (nikego@gmail.com)

#include <stdlib.h>
#include <string.h>
#include <FL/Fl.H>
#include <FL/x.H>


#define COLORVAL ulong
 
#define ARGB(a,r,g,b)	((COLORVAL)(((unsigned char)(r)|\
				(((unsigned)(unsigned char)(g))<<8))|\
				(((unsigned long)(unsigned char)(b))<<16)|\
				(((unsigned long)(unsigned char)(a))<<24)))
#define RGB(r,g,b)	ARGB(255,(r),(g),(b))		/* rgb full alpha*/
#define iRGB(r,g,b)	ARGB(0,(r),(g),(b))		/* rgb no alpha*/
#define GR_RGB(r,g,b)	RGB(r,g,b)

#define DoRed			(1<<0)
#define DoGreen			(1<<1)
#define DoBlue			(1<<2)

/*
 * Conversion from RGB to PIXELVAL
 */
#if 0
/* create 24 bit 8/8/8 format pixel (0x00RRGGBB) from RGB triplet*/
#define RGB2PIXEL888(r,g,b)	\
	(((r) << 16) | ((g) << 8) | (b))

/* create 16 bit 5/6/5 format pixel from RGB triplet */
#define RGB2PIXEL565(r,g,b)	\
	((((r) & 0xf8) << 8) | (((g) & 0xfc) << 3) | (((b) & 0xf8) >> 3))

/* create 16 bit 5/5/5 format pixel from RGB triplet */
#define RGB2PIXEL555(r,g,b)	\
	((((r) & 0xf8) << 7) | (((g) & 0xf8) << 2) | (((b) & 0xf8) >> 3))

/* create 8 bit 3/3/2 format pixel from RGB triplet*/
#define RGB2PIXEL332(r,g,b)	\
	(((r) & 0xe0) | (((g) & 0xe0) >> 3) | (((b) & 0xc0) >> 6))

#endif

/* pixel packing was taken from directfb's /src/gfx/convert.h */

#define PIXEL_RGB332(r,g,b)    ( (((r)&0xE0)     ) | \
                                 (((g)&0xE0) >> 3) | \
                                 (((b)&0xC0) >> 6) )

#define PIXEL_ARGB1555(a,r,g,b)( (((a)&0x80) << 8) | \
                                 (((r)&0xF8) << 7) | \
                                 (((g)&0xF8) << 2) | \
                                 (((b)&0xF8) >> 3) )

#define PIXEL_RGB555(r,g,b)  ( (((r)&0xF8) << 7) | \
                                 (((g)&0xF8) << 2) | \
                                 (((b)&0xF8) >> 3) )

#define PIXEL_ARGB2554(a,r,g,b)( (((a)&0xC0) << 8) | \
                                 (((r)&0xF8) << 6) | \
                                 (((g)&0xF8) << 1) | \
                                 (((b)&0xF0) >> 4) )

#define PIXEL_ARGB4444(a,r,g,b)( (((a)&0xF0) << 8) | \
                                 (((r)&0xF0) << 4) | \
                                 (((g)&0xF0)     ) | \
                                 (((b)&0xF0) >> 4) )

#define PIXEL_RGB444(r,g,b)  ( (((r)&0xF0) << 4) | \
                                 (((g)&0xF0)     ) | \
                                 (((b)&0xF0) >> 4) )

#define PIXEL_RGB16(r,g,b)     ( (((r)&0xF8) << 8) | \
                                 (((g)&0xFC) << 3) | \
                                 (((b)&0xF8) >> 3) )

#define PIXEL_RGB18(r,g,b)     ( (((r)&0xFC) << 10) | \
                                 (((g)&0xFC) <<  4) | \
                                 (((b)&0xFC) >>  2) )

#define PIXEL_RGB32(r,g,b)     ( ((r) << 16) | \
                                 ((g) <<  8) | \
                                  (b) )

#define PIXEL_ARGB(a,r,g,b)    ( ((a) << 24) | \
                                 ((r) << 16) | \
                                 ((g) << 8)  | \
                                  (b) )

#define PIXEL_ARGB1666(a,r,g,b) ( (((a)&0x80) << 11) | \
                                  (((r)&0xFC) << 10) | \
                                  (((g)&0xFC) <<  4) | \
                                  (((b)&0xFC) >>  2) )

#define PIXEL_ARGB6666(a,r,g,b) ( (((a)&0xFC) << 16) | \
                                  (((r)&0xFC) << 10) | \
                                  (((g)&0xFC) <<  4) | \
                                  (((b)&0xFC) >>  2) )

/*
 * Conversion from PIXELVAL to red, green or blue components
 */
/* return 8/8/8 bit r, g or b component of 24 bit pixelval*/
#define PIXEL888RED(pixelval)		(((pixelval) >> 16) & 0xff)
#define PIXEL888GREEN(pixelval)		(((pixelval) >> 8) & 0xff)
#define PIXEL888BLUE(pixelval)		((pixelval) & 0xff)

/* return 5/6/5 bit r, g or b component of 16 bit pixelval*/
#define PIXEL565RED(pixelval)		(((pixelval) >> 11) & 0x1f)
#define PIXEL565GREEN(pixelval)		(((pixelval) >> 5) & 0x3f)
#define PIXEL565BLUE(pixelval)		((pixelval) & 0x1f)

/* return 5/5/5 bit r, g or b component of 16 bit pixelval*/
#define PIXEL555RED(pixelval)		(((pixelval) >> 10) & 0x1f)
#define PIXEL555GREEN(pixelval)		(((pixelval) >> 5) & 0x1f)
#define PIXEL555BLUE(pixelval)		((pixelval) & 0x1f)

/* return 3/3/2 bit r, g or b component of 8 bit pixelval*/
#define PIXEL332RED(pixelval)		(((pixelval) >> 5) & 0x07)
#define PIXEL332GREEN(pixelval)		(((pixelval) >> 2) & 0x07)
#define PIXEL332BLUE(pixelval)		((pixelval) & 0x03)

/* return a pixel value's component r, g, b parts*/
void _nxPixel2RGB(Display* display, unsigned long color,
	   unsigned short *red, unsigned short *green, unsigned short *blue)
{
	/* the PIXELxxx macros below only return the actual color bits
	 * in the pixelval.  Thus, they need left shifting
	 * in order to become a normal r,g,b value.  In addition,
	 * they need to be shifted left another 8 bits to become
	 * X11 compatible colors.
	 */
/* nikita added - using palette  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
	int pix_f = pixel_format();
	if ((pix_f == DSPF_LUT8) || (pix_f == DSPF_ALUT44) )
	 {
		*blue   = PIXEL888RED(color)   << (8 + 0);
		*green  = PIXEL888GREEN(color) << (8 + 0);
		*red    = PIXEL888BLUE(color)  << (8 + 0);
	}
	else
	{ /*old code follow */
		switch (pix_f) {
		case DSPF_RGB332:
			*red    = PIXEL332RED(color)  << (8 + 5);
			*green  = PIXEL332GREEN(color)<< (8 + 5);
			*blue   = PIXEL332BLUE(color) << (8 + 6);
			break;
		case DSPF_ARGB1555:
			*red    = PIXEL555RED(color)  << (8 + 3);
			*green  = PIXEL555GREEN(color)<< (8 + 3);
			*blue   = PIXEL555BLUE(color) << (8 + 3);
			break;
		case DSPF_RGB16:
			*red    = PIXEL565RED(color)  << (8 + 3);
			*green  = PIXEL565GREEN(color)<< (8 + 2);
			*blue   = PIXEL565BLUE(color) << (8 + 3);;
			break;
		case DSPF_ARGB:  	
		case DSPF_RGB24:
		case DSPF_RGB32:  
			*red    = PIXEL888RED(color)  << (8 + 0);
			*green  = PIXEL888GREEN(color)<< (8 + 0);
			*blue   = PIXEL888BLUE(color) << (8 + 0);
 			break;
		default:
			break;		
		}
	}
}

int 
get_rgb_color(int xcolor, uchar& r, uchar& g, uchar& b){
	XColor xcol={0};
	xcol.pixel = xcolor;
	XQueryColor(fl_display, _nxDefaultColormap(fl_display), &xcol);
  	r = xcol.red    >> 8;
  	g = xcol.green  >> 8;
  	b = xcol.blue   >> 8;
  	return 0;	
}

int get_index_color(uchar r,uchar g,uchar b)
{
	XColor xcol;
	xcol.red    = r << 8;
	xcol.green  = g << 8;
	xcol.blue   = b << 8;
	XAllocColor(fl_display,_nxDefaultColormap(fl_display),&xcol);
	return xcol.pixel;
}

int XSetForeground(Display* display, GC gc, int color_index)
{
	if (gc) {
		//int pix_f = pixel_format();
		//if (pix_f == DSPF_LUT8){  		
		//	set_color(surface, color_index);
	 	//} else {
	 		uchar r,g,b;
	 		get_rgb_color(color_index, r, g ,b);
			set_color(gc,r,g,b,255);
	 	//}
	}
	return 0;
}

int XFreeColors(Display* display, Colormap colormap, unsigned long pixels[],
	    int npixels, unsigned long planes)
{
	int n;
	nxColormap *local = _nxFindColormap(colormap);

	if (!local)
		return 0;

	/* FIXME:  We should be smart and pack the array, blah, blah blah.... */
	/* I just stick this in so it will be easier later                    */

	for (n = 0; n < npixels; n++) {
		unsigned long index = pixels[n];

		if ((long)index >= local->cur_color || local->colorval[index].ref==0)
			continue;
		--local->colorval[index].ref;
	}

	return 1;
}


typedef struct hash_node{
	unsigned long rgb;
	int index;
	hash_node* next;
}hash_node_;

#define hash_ar_size 16384

#define mul_k 10368889

hash_node** hash_array = 0; 
#define hash_func(r, g, b) ((((r<<8)|g)<<8)|b)% 1001)
#define hash_func_2(rgb) ((ulong)(rgb*mul_k) >> 18)
#define hash_func_3(rgb) (rgb % 1001)

#define SQR(a) sqr(a)

// DirectFB's palette function SetEntries() is very slow (sometimes)
// we upload current colormap to the palette if it need

int
XFlushColormap(Display* display, Colormap colormap)
{
	if (!display || !fl_screen)
		return 0;
	int pix_f = pixel_format();
	if ((pix_f != DSPF_LUT8)&& (pix_f != DSPF_ALUT44)	)
		return 0;
		 
	nxColormap* local = _nxFindColormap(colormap);
	if (!local)
		return 0;
		
	int to_save = local->cur_color - local->saved_color; 
	if (to_save == 0)
		return 0;
		
	IDirectFBPalette* pal;
	get_palette(fl_screen, &pal);
	
	pal->SetEntries(pal, &local->colors[local->saved_color], to_save, local->saved_color); 
	pal->Release(pal);
	
	local->saved_color = local->cur_color;
	return to_save;
}
/*
 * Convert a color from RGB to pixel format.  If hw is non-palette
 * this routine will never allocate a colormap entry.  (FIXME?)
 * This works for now since we special case XPutImage for truecolor
 * hardware.
 */
int
XAllocColor(Display* display, Colormap colormap, XColor * in_out)
{
  if (!display || !fl_screen)
    return 0;

  int i;
  nxColormap * local;
  u32 rgb;
  u32 r = in_out->red   >> 8;
  u32 g = in_out->green >> 8;
  u32 b = in_out->blue  >> 8;

  int pix_f = pixel_format();
  if ((pix_f != DSPF_LUT8) && (pix_f != DSPF_ALUT44)){ //check for non-palette
    switch (pix_f){
      case DSPF_RGB332:
        in_out->pixel = PIXEL_RGB332( r, g, b );
        break;
      case DSPF_ARGB1555:
        in_out->pixel = PIXEL_ARGB1555( 0, r, g, b );
        break;
      //case DSPF_RGB555:
      //  in_out->pixel = PIXEL_RGB555( r, g, b );
      //  break;
      case DSPF_ARGB2554:
        in_out->pixel = PIXEL_ARGB2554( 0, r, g, b );
        break;
      case DSPF_ARGB4444:
        in_out->pixel = PIXEL_ARGB4444( 0, r, g, b );
        break;
      //case DSPF_RGB444:
      //  in_out->pixel = PIXEL_RGB444( r, g, b );
      //  break;
      case DSPF_RGB16:
        in_out->pixel = PIXEL_RGB16( r, g, b );
        break;
#if DIRECTFB_MAJOR_VERSION >= 1        
      case DSPF_RGB18:
      case DSPF_ARGB1666:
      case DSPF_ARGB6666:
        in_out->pixel = PIXEL_RGB18( r, g, b );
        break;
#endif
      case DSPF_RGB24:
      case DSPF_RGB32:
      case DSPF_ARGB:
      case DSPF_AiRGB:
        in_out->pixel = PIXEL_RGB32( r, g, b );
        break;
      default:
        in_out->pixel = 0;
    }
    return 1;
  }
  /* not truecolor hardware, allocate a colormap index*/
  rgb = GR_RGB(r, g, b);

  local = _nxFindColormap(colormap);
  if (!local)
    return 0;

  if (hash_array == 0){
    hash_array = (hash_node**)malloc(hash_ar_size* sizeof(hash_node*));
    memset(hash_array, 0, hash_ar_size* sizeof(hash_node*));
  }
  int hash_key = hash_func_2(rgb);
  hash_node* node = hash_array[hash_key],*tmp;
  while (node){
    if (node->rgb == rgb){
      i = node->index;
      local->colorval[i].ref++;
      in_out->pixel = i;
      return 1;
    }
    node = node->next;
  }
  tmp       = (hash_node*)malloc(sizeof(hash_node));
  tmp->rgb  = rgb;
  tmp->next = hash_array[hash_key];
  hash_array[hash_key]  = tmp;
  nxColorval* colors    = local->colorval;
  colors[local->cur_color].value = (long)rgb; // save the value in last position
  nxColorval min_color_val;
  int min_dist_color  = 10000000;
  int min_dist_idx    = -1,col_dist;
  for (i = 0; ;i++){
    if (colors->value == (long)rgb) {break;}
    // FindBestMatch() working very slow...
    // because we calculate a distinction between colors
    col_dist = SQR(colors->r - r) + SQR(colors->g - g) + SQR(colors->b - b);
    if (col_dist < min_dist_color){
      min_dist_color = col_dist; // save minimal distinction
      min_dist_idx = i;
    }
    colors++;
  }
  if (i < local->cur_color){ //if 'i' is not last value then we find the 'rgb'
    local->colorval[i].ref++;
    tmp->index = in_out->pixel = i;
    return 1;
  }
  if (local->cur_color >= 256){
    if (min_dist_idx == -1)//it's impossible...
      return 0;
    min_color_val = local->colorval[ min_dist_idx ];
    in_out->red = min_color_val.r << 8;
    in_out->green = min_color_val.g << 8;
    in_out->blue = min_color_val.b << 8;

    tmp->index = in_out->pixel = min_dist_idx;
    return 1;
  }
  /* Otherwise allocate a new item */
  if (i >= local->color_alloc){
    nxColorval *cv = (nxColorval *)Xrealloc(local->colorval,
      (local->color_alloc + 32) * sizeof(nxColorval));
    if (!cv)
      return 0;

    local->colorval = cv;
    local->color_alloc += 32;
  }
  tmp->index = in_out->pixel = local->cur_color;
  i = local->cur_color++;
  local->colorval[i].value = rgb;
  local->colorval[i].r = local->colors[i].r = r;
  local->colorval[i].g = local->colors[i].g = g;
  local->colorval[i].b = local->colors[i].b = b;
  local->colorval[i].ref = 1;
  if (PIXEL_RGB32(r, g, b) == (BG_COLOR_RGB32 & 0xFFFFFF)){
    _TRACE("transparent color %x = %d\n", rgb, i);
    local->colors[i].a = 0;
  }else
    local->colors[i].a = 255;

  //save the colormap to hardware will be done via calling XFlushColormap()
  return 1;
}
/*
 * Local colormap functions to emulate X11 colormap scheme
 */

static nxColormap *colormap_hash[32];
static int colormap_id = 1;
static Colormap _defaultColormap = 0;

static Colormap createColormap(int color_count)
{
	int val = (colormap_id - 1) % 32;
	nxColormap *local = (nxColormap *)Xcalloc(sizeof(nxColormap), 1);

	if (!local)
		return 0;
	local->id = colormap_id++;

	if (color_count) {
		/* The initial count of colors is simply a good suggestion.
		 * It can grow / shrink
		 */
		local->colorval =
			(nxColorval *)Xmalloc(color_count *
						sizeof(nxColorval));

		if (local->colorval) {
			local->color_alloc = color_count;
			local->cur_color = 0;
			local->saved_color= 0;
			memset(local->colors, 0, sizeof(local->colors));
		}
	}

	if (!colormap_hash[val])
		colormap_hash[val] = local;
	else {
		nxColormap *t = colormap_hash[val];
		while (t->next)
			t = t->next;
		t->next = local;
	}

	return local->id;
}

#if LATER
static void
destroyColormap(Colormap id)
{
	int val = id % 32;
	nxColormap *ptr, *prev;

	for (ptr = colormap_hash[val], prev = 0; ptr;
	     prev = ptr, ptr = ptr->next) {
		if (ptr->id == id) {
			if (prev)
				prev->next = ptr->next;
			else
				colormap_hash[val] = ptr->next;

			if (ptr->colorval)
				Xfree(ptr->colorval);
			Xfree(ptr);
			return;
		}
	}
}
#endif

nxColormap * _nxFindColormap(Colormap id)
{
	int val = (id - 1) % 32;
	nxColormap *ptr = colormap_hash[val];

	if (id == 0)
		return 0;

	for (ptr = colormap_hash[val]; ptr; ptr = ptr->next)
		if (ptr->id == (long)id)
			return ptr;

	return 0;
}

/* Populate the colormap with the 32 system colors */
Colormap _nxDefaultColormap(Display* display)
{
	XColor t;
	Colormap cm;

	if (_defaultColormap)
		return _defaultColormap;

/* nikita: add loading palette*/

	cm = createColormap(256 + 1);
#if 0	
	IDirectFBPalette* pal;
	surface->GetPalette(surface, &pal);
	unsigned pal_size;
	pal->GetSize(pal, &pal_size);
	DFBColor col[256];
	pal->GetEntries(pal, col, pal_size < 256 ? pal_size : 256 , 0);
	pal->Release(pal);
	int k = 0;
	for (k = 0; k < pal_size; k++)
	{
		t.red = col[k].r << 8;
		t.blue = col[k].b << 8;
		t.green = col[k].g << 8; 
		t.pixel = k;
		if (k < 32)
			printf("%d = (%x,%x,%x)\n", k, t.red, t.green, t.blue);
		XAllocColor(surface, cm, &t);
		if (k < 32)
			printf("%d = (%x,%x,%x)\n", k, t.red, t.green, t.blue);
	}
#endif
	/*
	 * Note the following won't allocate a colormap entry
	 * on truecolor hardware...
	 */

	/* Create the 0th index to have black */
  	t.red = t.blue = t.green = 0 << 8; 
	XAllocColor(display, cm, &t);

	/* Create the 1st index to have white */
	t.red = t.blue = t.green = 255 << 8;
	XAllocColor(display, cm, &t);

	_defaultColormap = cm;
	
	return cm;
}

/*
 * Create a colormap.  We always use the default colormap
 * for the time being.  Non-default visuals aren't supported either.
 */
Colormap XCreateColormap(Display * display, Window w, Visual * visual, int alloc)
{
	return _nxDefaultColormap(display);
#if 0 // FIXME?
	switch (visual->class) {
	case DirectColor:
		return createColormap(0);

	case GrayScale:
	case PseudoColor:
		return createColormap(visual->map_entries);

	case StaticGray:
	case StaticColor:
	case TrueColor:
		if (alloc == AllocAll)
			return 0;
		return createColormap(visual->map_entries);
	}

	return 0;
#endif
}

int XFreeColormap(GC gc, Colormap colormap)
{
#if 0 // FIXME?
	destroyColormap(colormap);
#endif
	/*for(int i = 0; i< hash_ar_size;i++){ 
				hash_node* node = hash_array[i],*tmp;
    			if (node){
    				do{
    					tmp = node->next;
    					free(node);
    					node = 0;
    					node = tmp;
    				}while(node);
    			}
			}
			free(hash_array);
			hash_array = 0;*/
	return 1;
}

/* convert an X11 pixelval to Microwindows colorval*/
GR_COLOR _nxColorvalFromPixelval(Display* display, unsigned long pixelval)
{
	XColor c;

	/* convert pixelval to colorval*/
	c.pixel = pixelval;
	XQueryColor(display, _nxDefaultColormap(display), &c);

	return GR_RGB(c.red >> 8, c.green >> 8, c.blue >> 8);
}

static int
queryColor(Display * display, nxColormap * local, XColor * def)
{
	if (def->pixel >= local->cur_color)
		return 0;

	_nxPixel2RGB(display, local->colorval[def->pixel].value, &def->red,
		   &def->green, &def->blue);

	def->flags |= (DoRed | DoGreen | DoBlue);
	return 1;
}

int XQueryColor(Display* display, Colormap colormap, XColor * def)
{
	nxColormap *local;
	int pix_f = pixel_format();
	
	if ((pix_f != DSPF_LUT8) && (pix_f != DSPF_ALUT44)) {
		_nxPixel2RGB(display, def->pixel, &def->red, &def->green,
			   &def->blue);
		return 1;
	}

	if (!(local = _nxFindColormap(colormap)))
		return 0;

	return queryColor(display, local, def);
}

int XQueryColors(Display* display, Colormap colormap, XColor * def_in,
	     int ncolors)
{
	nxColormap *local;
	int i;
	int pix_f = pixel_format();
	
	if ((pix_f != DSPF_LUT8) && (pix_f != DSPF_ALUT44)) {
		for (i = 0; i < ncolors; i++)
			_nxPixel2RGB(display, def_in[i].pixel, &def_in[i].red,
				   &def_in[i].green, &def_in[i].blue);

		return 1;
	}

	if (!(local = _nxFindColormap(colormap)))
		return 0;

	for (i = 0; i < ncolors; i++)
		queryColor(display, local, &def_in[i]);

	return 1;
}

#define ISBLANK(c)	((c) == ' ' || (c) == '\t')
#define ISDIGIT(c)	((c) >= '0' && (c) <= '9')
#define ISEOL(c)	((c) == '\r' || (c) == '\n')

/* get next int from buffer, return next buf position*/
static char *
strInt(int *retint, char *buf)
{
	char *	q;
	char	tmp[256];

	while (ISBLANK(*buf))
		++buf;
	q = tmp;
	while (ISDIGIT(*buf))
		*q++ = *buf++;
	*q = 0;
	*retint = atoi(tmp);
	return buf;
}

/* return string to end of line from buffer*/
static void
strEol(char *retbuf, char *buf)
{
	while (ISBLANK(*buf))
		++buf;
	while (!ISEOL(*buf))
		*retbuf++ = *buf++;
	*retbuf = 0;
}

GR_COLOR
GrGetColorByName(char *colorname, int *retr, int *retg, int *retb)
{
	FILE *fp;
	int r = 0, g = 0, b = 0;
	char buf[256];

	fp = fopen(X11_RGBTXT, "r");
	if (!fp)
		return 0;

	while (fgets(buf, 256, fp) != NULL) {
		if (buf[0] != '!') {
			char *p;
			char name[256];
	
			p = strInt(&r, buf);
			p = strInt(&g, p);
			p = strInt(&b, p);
			strEol(name, p);

			if (strcasecmp(name, colorname) == 0) {
				if (retr)
					*retr = r;
				if (retg)
					*retg = g;
				if (retb)
					*retb = b;
				return GR_RGB(r, g, b);
			}
		}
	}
	fclose(fp);

	return 0;
}


/* Parse colors of format:
 * #RGB #RRGGBB #RRRGGGBBB  #RRRRGGGGBBBB
 */
static int
_parseColorStr(_Xconst char **str, int size)
{
	char parse[5];
	unsigned long val;

	strncpy(parse, *str, size);
	parse[size + 1] = '\0';
	val = strtol(parse, 0, 16);
	*str += size;
	return (val);
}

Status
XParseColor(Display * display, Colormap colormap, _Xconst char *spec,
	XColor *exact)
{
	int r, g, b;

	/* This is the new and preferred way */
	if (strncmp(spec, "rgb:", 4) == 0) {
		sscanf(spec + 4, "%x/%x/%x", &r, &g, &b);
	} else {
		if (spec[0] != '#') {
			/* try to parse the color name */
			if (GrGetColorByName((char *) spec, &r, &g, &b) == 0) {
				printf("XParseColor: bad parse on %s\n", spec);
				return 0;
			}
		} else {
			_Xconst char *p = spec + 1;
			unsigned long val;

			switch (strlen(p)) {
			case 3:
				r = _parseColorStr(&p, 1);
				g = _parseColorStr(&p, 1);
				b = _parseColorStr(&p, 1);
				break;

			case 6:
				val = strtol(p, 0, 16);
				r = (val >> 16) & 0xFF;
				g = (val >> 8) & 0xFF;
				b = (val & 0xFF);
				break;

			case 12:	/* #RRRGGGBBB */
				r = _parseColorStr(&p, 3);
				g = _parseColorStr(&p, 3);
				b = _parseColorStr(&p, 3);

				if (r > 0xFF)
					r >>= 4;
				if (g > 0xFF)
					g >>= 4;
				if (b > 0xFF)
					b >>= 4;

				break;

			default:
				printf("XParseColor: invalid size %d on %s\n",
				       (int)strlen(p), p);
				return 0;
			}
		}
	}

	exact->red = r << 8;
	exact->green = g << 8;
	exact->blue = b << 8;
	exact->flags |= (DoRed | DoGreen | DoBlue);

	return 1;
}

Status
XLookupColor(Display * display, Colormap colormap, _Xconst char *spec,
	     XColor * exact, XColor * screen)
{
	Status stat = XParseColor(display, colormap, spec, exact);

	if (!stat)
		return stat;

	/* FIXME:  Should do a system look up for the right color */
	/* This will come back and haunt you on palettized machines */

	screen->red = exact->red;
	screen->green = exact->green;
	screen->blue = exact->blue;

	screen->flags |= (DoRed | DoGreen | DoBlue);

	return stat;
}

