//
// "$Id: _listfonts.cxx,v 1.1 2006-04-06 14:01:27 nikego Exp $"
//
// Font-specific code for the Fast Light Tool Kit (FLTK).
//
// Copyright 1998-2005 by Bill Spitzak and others.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems on the following page:
//
//     http://www.fltk.org/str.php
//
// This file contains font-specific code for fltk which is always linked
// in.  Search other files for "DIRECTFB" or filenames ending in _directfb.cxx
// for other system-specific code.
//
// This file was taken from  Microwindows (Nano-X Window System) project
// that is a part of the nxlib project.
// http://www.microwindows.org
// Edited by Nikita Egorov (nikego@gmail.com) 

char **_nxfontlist = 0;
int _nxfontcount = 0;

//using namespace std;

FILE *
_nxLoadFontDir(char *str)
{
	char path[256];

	sprintf(path, "%s/fonts.dir", str);
	return fopen(path, "r");
}


void
_nxSetDefaultFontDir(void)
{
	int i;

	if (_nxfontlist) {
		for (i = 0; i < _nxfontcount; i++)
			Xfree(_nxfontlist[i]);
		Xfree(_nxfontlist);
	}

	_nxfontlist = (char **) Xcalloc(2, sizeof(char *));
	_nxfontlist[0] = strdup(X11_FONT_DIR1);
	_nxfontlist[1] = strdup(X11_FONT_DIR2);
	_nxfontcount = 2;
}

struct _list {
	char **list;
	int alloc;
	int used;
	struct _list *next;
};

static struct _list *g_fontlist = 0;

static struct _list *
_createFontList(void)
{
	struct _list *ptr;

	if (!g_fontlist)
		ptr = g_fontlist =
			(struct _list *)Xcalloc(sizeof(struct _list), 1);
	else {
		struct _list *t;
		for (t = g_fontlist; t->next; t = t->next)
			continue;
		ptr = t->next =
			(struct _list *)Xcalloc(sizeof(struct _list), 1);
	}
	return ptr;
}

static int
_addFontToList(struct _list *list, char *font)
{
	if (list->alloc == 0) {
		list->list = (char**)Xmalloc(5 * sizeof(char *));
		list->alloc = 5;
	} else if (list->used == list->alloc) {
		list->list = (char**)Xrealloc(list->list,
				(list->alloc + 5) * sizeof(char *));
		list->alloc += 5;
	}

	list->list[list->used++] = strdup(font);
	return list->used;
}

static char **
_getFontList(struct _list *list, int *size)
{
	if (!list->list) {
		*size = 0;
		return 0;
	}

	if (list->alloc != list->used)
		list->list =
			(char**)Xrealloc(list->list, (list->used) * sizeof(char *));

	*size = list->used;
	return list->list;
}

static void
_freeFontList(char **fontlist)
{
	struct _list *ptr = g_fontlist;
	struct _list *prev = 0;

	if (!fontlist)
		return;

	while (ptr) {
		if (ptr->list == fontlist) {
			int i;
			for (i = 0; i < ptr->used; i++)
				Xfree(ptr->list[i]);

			Xfree(ptr->list);

			if (ptr == g_fontlist)
				g_fontlist = ptr->next;
			else
				prev->next = ptr->next;

			return;
		}
		prev = ptr;
		ptr = ptr->next;
	}
}
#if 1
static int
_dashcount(char *name)
{
	int	ndashes = 0;

	while (*name)
		if (*name++ == '-')
			++ndashes;
	return ndashes;
}

static int
patternmatch(char *pat, int patdashes, char *string, int stringdashes)
{
	int c, t;

	if (stringdashes < patdashes)
		return 0;

	for (;;) {
	    switch (c = *pat++) {
	    case '*':
		if (!(c = *pat++))
			return 1;
		if (c == '-') {
			patdashes--;
			for (;;) {
				while ((t = *string++) != '-')
					if (!t)
						return 0;
				stringdashes--;
				if (patternmatch(pat, patdashes, string, stringdashes))
					return 1;
				if (stringdashes == patdashes)
					return 0;
			}
		} else {
			for (;;) {
				while ((t = *string++) != c) {
					if (!t)
						return 0;
					if (t == '-') {
						if (stringdashes-- < patdashes)
							return 0;
					}
				}
				if (patternmatch(pat, patdashes, string, stringdashes))
					return 1;
			}
		}
	    case '?':
		if (*string++ == '-')
			stringdashes--;
		break;
	    case '\0':
		return (*string == '\0');
	    case '-':
		if (*string++ == '-') {
			patdashes--;
			stringdashes--;
			break;
		}
		return 0;
	    default:
		if (c == *string++)
			break;
		return 0;
	    }
	}
}

static int
match(char *pat, char *string)
{
	return patternmatch(pat, _dashcount(pat), string, _dashcount(string));
}

#else
static int
match(char *pattern, char *font)
{
	char *src = pattern;
	char *dst = font;
	int m = 0;

	if (!strcmp(pattern, "-*"))
		pattern = "*"; //"-*-*-*-*-*-*-*-*-*-*-*-*-*-*";

	if (!strcmp(pattern, "*"))
		return 1;

	while (1) {
		char *s, *e;

		src++;
		dst++;

		/* Skip any wildcardish stuff */
		if (*src == '*') {
			src++;
			if (!*src)
				break;

			for (; *dst && *dst != '-'; dst++)
				continue;
			if (!*dst)
				break;

			continue;
		}

		/* Find the end of the compare */
		s = src;
		e = src;
		for (; *e && *e != '-'; e++)
			continue;
		_TRACE("strnicmp '%s' '%s'\n", s ,dst);
		if (strnicmp(s, dst, (int) (e - s)) == 0) {
			m = 1;
		} else {
			m = 0;
			break;
		}

		src += (int) (e - s);
		dst += (int) (e - s);
	}
	return m;
}
#endif

static char **
_findFontPattern(char *_pattern, int maxnames, int *count)
{
	char pattern[128];
	strcpy(pattern, _pattern);
	strlwr(pattern);
	//_TRACE("pattern = '%s'\n", pattern);
	
	struct _list *flist = _createFontList();
	int fcount = 0, i = 0, f;
	char buffer[128];

	if (!_nxfontlist)
		_nxSetDefaultFontDir();

	for (f = 0; f < _nxfontcount; f++) {
		//_TRACE("trying open font list %s\n", _nxfontlist[f]);
		FILE *fontdir = _nxLoadFontDir(_nxfontlist[f]);
		if (!fontdir)
			continue;

		fgets(buffer, 128, fontdir);
		fcount = atoi(buffer);

		if (!fcount) {
			fclose(fontdir);
			continue;
		}

		for (i = 0; i < fcount; i++) {
			char *font;

			memset(buffer, 0, 128);
			fgets(buffer, 128, fontdir);

			/* Remove the end 'o line */
			buffer[strlen(buffer) - 1] = '\0';

			/* Find the field seperator */
			font = strchr(buffer, ' ');
			*font++ = '\0';
			strlwr(font);

			if (match(pattern, font))
				if (_addFontToList(flist, font) == maxnames)
					break;
		}
		fclose(fontdir);
	}
	return _getFontList(flist, count);
}

/*
 * Compare two strings just like strcmp, but preserve decimal integer
 * sorting order, i.e. "2" < "10" or "iso8859-2" < "iso8859-10" <
 * "iso10646-1". Strings are sorted as if sequences of digits were
 * prefixed by a length indicator (i.e., does not ignore leading zeroes).
 *
 * Markus Kuhn <Markus.Kuhn@cl.cam.ac.uk>
 */
#define isdigit(c) ('0' <= (c) && (c) <= '9')
static int
strcmpn(unsigned char *s1, unsigned char *s2)
{
	int digits, predigits = 0;
	unsigned char *ss1, *ss2;

	while (1) {
		if (*s1 == 0 && *s2 == 0)
			return 0;
		digits = isdigit(*s1) && isdigit(*s2);
		if (digits && !predigits) {
			ss1 = s1;
			ss2 = s2;
			while (isdigit(*ss1) && isdigit(*ss2))
				ss1++, ss2++;
			if (!isdigit(*ss1) && isdigit(*ss2))
				return -1;
			if (isdigit(*ss1) && !isdigit(*ss2))
				return 1;
		}
		if (*s1 < *s2)
			return -1;
		if (*s1 > *s2)
			return 1;
		predigits = digits;
		s1++, s2++;
	}
}

static int
comparefunc(const void* a, const void* b)
{
	unsigned char *aa = *(unsigned char **)a;
	unsigned char *bb = *(unsigned char **)b;

	return strcmpn(aa, bb);
}

/***************************************************************************/
#if 1
char **
_nxGetFontDir(int *count)
{
	int i;

	int size;
	char **ret;

	if (!_nxfontcount)
		_nxSetDefaultFontDir();

	size = _nxfontcount + 1;
	ret = (char **) Xcalloc(size, sizeof(char *));

	for (i = 0; i < _nxfontcount; i++)
		ret[i] = strdup(_nxfontlist[i]);
	ret[_nxfontcount] = 0;

	*count = _nxfontcount;
	return (ret);
}

void
_nxSetFontDir(char **directories, int ndirs)
{
	int i;

	if (_nxfontlist) {
		for (i = 0; i < _nxfontcount; i++)
			Xfree(_nxfontlist[i]);
		Xfree(_nxfontlist);
	}

	_nxfontlist = (char **) Xcalloc(ndirs, sizeof(char *));
	for (i = 0; i < ndirs; i++)
		_nxfontlist[i] = strdup(directories[i]);

	_nxfontcount = ndirs;
}

void
_nxFreeFontDir(char **list)
{
	int i;

	if (list) {
		for (i = 0; list[i]; i++)
			Xfree(list[i]);
		Xfree(list);
	}
}
// end of namespace list_fonts

char **
XListFonts(Display * display, const char *pattern, int maxnames,
	int *actual_count_return)
{
	int count;
	char **ret;

	ret = _findFontPattern((char *)pattern, maxnames, &count);
	*actual_count_return = count;

	/* sort the return, helps for lack of locale info at end of XLFD*/
	qsort((char *)ret, count, sizeof(char *), comparefunc);

	return ret;
}


int
XFreeFontNames(char **list)
{
	_freeFontList(list);
	return 1;
}
int
XSetFontPath(Display * display, char **directories, int ndirs)
{
	_nxSetFontDir(directories, ndirs);
	return 1;
}

char **
XGetFontPath(Display * display, int *npaths_return)
{
	return _nxGetFontDir(npaths_return);
}

int
XFreeFontPath(char **list)
{
	_nxFreeFontDir(list);
	return 1;
}
#else
Bool
XGetFontProperty(XFontStruct * font, Atom atom, unsigned long *value_return)
{
//_TRACE("XGetFontProperty called\n");
	switch (atom) {
	case XA_FONT:			/* 18*/
	case XA_UNDERLINE_POSITION:	/* 51*/
	case XA_UNDERLINE_THICKNESS:	/* 52*/
		break;
	default:
	//	_TRACE("XGetFontProperty: Unknown FontProperty Atom %d\n",
	//		(int)atom);
	}
	return 0;
}
#endif
