//
// "$Id: _loadfont.cxx,v 1.2 2006-10-31 09:31:14 nikego Exp $"
//
// Font loading code for the Fast Light Tool Kit (FLTK).
//
// Copyright 1998-2005 by Bill Spitzak and others.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems on the following page:
//
//     http://www.fltk.org/str.php
//
// This file contains font specific code for fltk which is always linked
// in.  Search other files for "DIRECTFB" or filenames ending in _directfb.cxx
// for other system-specific code.
//
// This file was taken from  Microwindows (Nano-X Window System) project
// that is a part of the nxlib project. http://www.microwindows.org
// Edited by Nikita Egorov (nikego@gmail.com)

static int
prefix(const char *prestr, char *allstr)
{
	while (*prestr)
		if (*prestr++ != *allstr++)
			return 0;
	if (*allstr != '.')
		return 0;
	return 1;
}

char *
_nxFindX11Font(const char *_xfontname)
{
	char xfontname[128]={0};
	strcpy(xfontname, _xfontname);
	strlwr(xfontname);
	int fcount, i, f;
	char *ret;
	char buffer[128];
/******************************************************************/
/* Added by Nikita */
/* Prepare for truetype fonts - deleting font size */
	char ttfname[128];
	strcpy(ttfname, xfontname);
	char* pos = strstr(ttfname, "--");
	if (pos) {
		pos += 2;
		*pos = '0';
		pos++;
		if (*pos != '-'){
			if ( pos[1] == '-')
				strcpy(pos, pos + 1);
		}
	}
	//printf("ttfname '%s' '%s'\n", ttfname, xfontname);
/******************************************************************/
	if (!_nxfontcount)
		_nxSetDefaultFontDir();

	/* Go through all of the font dirs */
	for (f = 0; f < _nxfontcount; f++) {
		FILE *fontdir = _nxLoadFontDir(_nxfontlist[f]);
		if (!fontdir)
			continue;

		fgets(buffer, 128, fontdir);
		fcount = atoi(buffer);

		if (!fcount) {
			fclose(fontdir);
			continue;
		}

		for (i = 0; i < fcount; i++) {
			char *file = buffer, *font = 0;

			fgets(buffer, 128, fontdir);

			/* Remove the end 'o line */
			buffer[strlen(buffer) - 1] = '\0';
			
			/******************************************************************/
			/* Added by Nikita */
			strlwr(buffer);
			const char* xfontname_ = xfontname; 
			if (strstr(buffer, ".ttf"))//checking for truetype font
			{
				xfontname_ = ttfname;				
			}
			/******************************************************************/

			/* Find the field seperator */

			font = strchr(buffer, ' ');
			*font++ = '\0';

			if (strcmp(xfontname_, font) == 0) {
				ret = (char *) Xmalloc(strlen(_nxfontlist[f]) +
						      strlen(file) + 2);
				sprintf(ret, "%s/%s", _nxfontlist[f], file);

				fclose(fontdir);
				return ret;
			}
		}

		/* not found, try <prefix.pcf> */
		fseek(fontdir, 0L, SEEK_SET);
		fgets(buffer, 128, fontdir);
		for (i = 0; i < fcount; i++) {
			char *file = buffer, *font = 0;

			fgets(buffer, 128, fontdir);

			/* Remove the end 'o line */
			buffer[strlen(buffer) - 1] = '\0';
			strlwr(buffer);

			/* Find the field seperator */

			font = strchr(buffer, ' ');
			*font++ = '\0';

			if (prefix(xfontname, file)) {
				ret = (char *) Xmalloc(strlen(_nxfontlist[f]) +
						      strlen(file) + 2);
				sprintf(ret, "%s/%s", _nxfontlist[f], file);

				fclose(fontdir);
				return ret;
			}
		}

		if (fontdir)
			fclose(fontdir);
	}
	return 0;
}

static int
any(int c, const char *str)
{
	while (*str)
		if (*str++ == c)
			return 1;
	return 0;
}

XFont
XLoadFont(Display* display , const char *name)
{
	XFont fbfont = 0;
	char *fontname=0;

	/* first check for wildcards*/
	if (any('*', name) || any('?', name)) {
		char **fontlist;
		int count;

		/* pick first sorted return value for now...*/
		fontlist = XListFonts(display, name, 100000, &count);
		if (fontlist)
			fontname = fontlist[0];
	} else
		fontname = (char *)name;

	/* first try to find from X11/fonts.dir file*/
	fontname = _nxFindX11Font(fontname);
	
	/* if not found, try 6x13 for "fixed"*/
	if (!fontname /*&& !strcmp(name, "fixed")*/)
    {// use default font of DirectFB
		//fontname = _nxFindX11Font("6x13");
        char defaultFont[PATH_MAX];
        snprintf(defaultFont, PATH_MAX, "%s/fonts/DroidSans.ttf", getenv("VA_RUNTIME_ROOT"));
		const char *sizepos = strstr(name, "--");
        DFBFontDescription font_dsc;
        font_dsc.flags = (DFBFontDescriptionFlags)(DFDESC_HEIGHT);
        font_dsc.height = 10;
		if (sizepos){
			sizepos += 2;
			font_dsc.height = atoi(sizepos);
		}
        create_font(defaultFont, &font_dsc, &fbfont);
    }

	/* found font, load into server*/
	if (fontname)
    {
		const char *sizepos = strstr(name, "--");
		DFBFontDescription font_dsc;	
		font_dsc.flags = (DFBFontDescriptionFlags)(DFDESC_HEIGHT/* |DFDESC_INDEX*/);
//        if (have_lut()){
//		  font_dsc.attributes = (DFBFontAttributes)(DFFA_MONOCHROME);
//          font_dsc.flags = (DFBFontDescriptionFlags)(font_dsc.flags | DFDESC_ATTRIBUTES) ;
//        }
		int size = 10;
		if (sizepos){
			sizepos += 2;
			size = atoi(sizepos);
		}
		/*font_dsc.index = 3; //sometimes it worked...*/
		font_dsc.height = size;
        create_font(fontname, &font_dsc, &fbfont);
	}else{
		_TRACE("font file not found\n");
	}
    if (fontname)
		Xfree(fontname);
		
	return fbfont;
}
