
#include "fl_directfb_eventfilter.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <signal.h>

class Fl_DirectFB_EventFilter_IVI : public Fl_DirectFB_EventFilter {
public:
    Fl_DirectFB_EventFilter_IVI()
        : mKeyAltL(false), mKeyCtrlL(false), mKeyShiftL(false)
    {
        const char* name = "NEED_MAP_AK2ME";
		char* value = getenv(name);
        if (value && (!strcasecmp(value, "true") || !strcasecmp(value, "on")))
            mNeedMapAK2ME = true;
        else
            mNeedMapAK2ME = false;
    }

    bool eventFilter(DFBEvent& e);

    void doPreProcess(DFBWindowEvent& event);

private:
    bool mNeedMapAK2ME;
    bool mKeyAltL;
    bool mKeyCtrlL;
    bool mKeyShiftL;

    int	m_alt_f;//Now, just for exit app use.
};

Fl_DirectFB_EventFilter* Fl_DirectFB_EventFilter::createEventFilter(const EventFilterType& type)
{
    if (type == InterVideo)
        return new Fl_DirectFB_EventFilter_IVI();
    assert(0);
    return NULL;
}

void Fl_DirectFB_EventFilter_IVI::doPreProcess(DFBWindowEvent& event)
{
    switch(event.key_id)
    {
        case DIKI_ALT_L:
            {
                m_alt_f = 1;
                break;
            }
        case DIKI_F4:
            {
                if(m_alt_f)
                {
                    m_alt_f = 0;
                    event.key_symbol = DIKS_ESCAPE;
                }
                break;
            }
        default:
            break;   		
    }
    //If it is not DIKI_ALT_L, reset it
    if(event.key_id != DIKI_ALT_L)
        m_alt_f = 0;
}

bool Fl_DirectFB_EventFilter_IVI::eventFilter(DFBEvent& e)
{
    if (e.clazz != DFEC_WINDOW) //FIXME
        return true;

    DFBWindowEvent& event = (DFBWindowEvent&)e;

    if (event.type == DWET_MOTION) {
        //printf("%s DWET_MOTION {%d, %d}\n", __FUNCTION__, event.cx, event.cy);
    }
        
    if (event.type != DWET_KEYDOWN)
        return true;

    doPreProcess(event);

    if (event.key_symbol == DIKS_BACK) {
        /// Map BACK to Shift+Tab
        event.modifiers = DIMM_SHIFT;
        event.key_id = DIKI_TAB;
        event.key_symbol = DIKS_TAB;
        return true;
    } else if (event.key_symbol == DIKS_POWER || event.key_symbol == DIKS_ESCAPE) {
        kill(0, SIGINT);
        return false;
    }

    if (mNeedMapAK2ME) {
        switch (event.key_id) {
            /// Map Arrow keys to F9/10/11/12 to Mouse emulator
            /// for focus navigation 
            case DIKI_UP:
                event.key_id = DIKI_F9;
                event.key_symbol = DIKS_F9;
                break;
            case DIKI_DOWN:
                event.key_id = DIKI_F10;
                event.key_symbol = DIKS_F10;
                break;
            case DIKI_LEFT:
                event.key_id = DIKI_F11;
                event.key_symbol = DIKS_F11;
                break;
            case DIKI_RIGHT:
                event.key_id = DIKI_F12;
                event.key_symbol = DIKS_F12;
                break;
            default:
                break;
        }//switch(){}
    }//if

    switch (event.key_id) {
        /// Map PG Up/Down to arrow keys
        case DIKI_PAGE_UP:
            // event.key_id = DIKI_UP;
            // event.key_symbol = DIKS_CURSOR_UP; 
            event.key_symbol = DIKS_SMALL_U; 
            break;
        case DIKI_PAGE_DOWN:
            // event.key_id = DIKI_DOWN;
            // event.key_symbol = DIKS_CURSOR_DOWN;
            event.key_symbol = DIKS_SMALL_D;
            break;

            /// Map HOME on RC for F5 (Refresh)
        ///case DIKI_HOME:
        ///    event.key_id = DIKI_F5;
        ///    event.key_symbol = DIKS_F5;
        ///    break;

            /// Map for Zoom/History keys
        case DIKI_ALT_L:
            mKeyAltL = true;
            return false;
        case DIKI_CONTROL_L:
            mKeyCtrlL = true;
            return false;
        case DIKI_SHIFT_L:          
            mKeyShiftL = true;
            return false;
        case DIKI_1:
            if (mKeyAltL && mKeyCtrlL && mKeyShiftL) 
            {
                event.key_symbol = DIKS_SMALL_H;
                mKeyCtrlL = false;
                mKeyCtrlL = false;
                mKeyAltL = false;
            }
            break;
        case DIKI_B:
            if (mKeyCtrlL) {
                if (mKeyShiftL) {
                    event.key_id = DIKI_F3;
                    event.key_symbol = DIKS_F3;
                } else { 
                    event.key_symbol = DIKS_F1;
                    event.key_id = DIKI_F1;
                }

                mKeyCtrlL = false;
                mKeyShiftL = false;
            }
            break;
        case DIKI_F:
            if (mKeyCtrlL) {
                if (mKeyShiftL) {
                    event.key_id = DIKI_F4;
                    event.key_symbol = DIKS_F4;
                } else { 
                    event.key_symbol = DIKS_F2;
                    event.key_id = DIKI_F2;
                }

                mKeyCtrlL = false;
                mKeyShiftL = false;
            }
            break;

            /// Map for Play/Pause and Stop functional keys
        case DIKI_P:
            if (mKeyCtrlL) 
            {
                event.key_symbol = DIKS_PLAYPAUSE;
                mKeyCtrlL = false;
            }
            break;
        case DIKI_S:
            if (mKeyCtrlL) 
            {
                event.key_symbol = DIKS_STOP;
            }
            mKeyCtrlL = false;
        case DIKI_R:
            if (mKeyCtrlL) 
            {
                event.key_symbol = DIKS_MUTE;
            }
            mKeyCtrlL = false;
            break;
        default:
            break;
    }
    return true;
}

